@@ -16,7 +16,6 @@ use std::borrow::Cow;
 
 #[cfg(any(feature = "parquet", feature = "csv-file", feature = "ipc"))]
 use polars_core::datatypes::PlHashMap;
-use polars_core::frame::groupby::DynamicGroupOptions;
 use polars_core::frame::hash_join::JoinType;
 use polars_core::prelude::*;
 #[cfg(feature = "dtype-categorical")]
@@ -32,15 +31,18 @@ use crate::logical_plan::optimizer::{
     predicate_pushdown::PredicatePushDown, projection_pushdown::ProjectionPushDown,
 };
 use crate::physical_plan::state::ExecutionState;
+
 #[cfg(any(feature = "parquet", feature = "csv-file"))]
 use crate::prelude::aggregate_scan_projections::agg_projection;
-use crate::prelude::drop_nulls::ReplaceDropNulls;
-use crate::prelude::fast_projection::FastProjection;
-use crate::prelude::simplify_expr::SimplifyBooleanRule;
-use crate::prelude::slice_pushdown::SlicePushDown;
+use crate::prelude::{
+    drop_nulls::ReplaceDropNulls, fast_projection::FastProjection,
+    simplify_expr::SimplifyBooleanRule, slice_pushdown::SlicePushDown, *,
+};
+
+use crate::logical_plan::FETCH_ROWS;
 use crate::utils::{combine_predicates_expr, expr_to_root_column_names};
-use crate::{logical_plan::FETCH_ROWS, prelude::*};
 use polars_arrow::prelude::QuantileInterpolOptions;
+use polars_io::RowCount;
 
 #[derive(Clone, Debug)]
 pub struct JoinOptions {
@@ -48,8 +50,6 @@ pub struct JoinOptions {
     pub force_parallel: bool,
     pub how: JoinType,
     pub suffix: Cow<'static, str>,
-    pub asof_by_left: Vec<String>,
-    pub asof_by_right: Vec<String>,
 }
 
 impl Default for JoinOptions {
@@ -59,8 +59,6 @@ impl Default for JoinOptions {
             force_parallel: false,
             how: JoinType::Left,
             suffix: "_right".into(),
-            asof_by_left: vec![],
-            asof_by_right: vec![],
         }
     }
 }
@@ -154,7 +152,7 @@ impl LazyFrame {
     pub(crate) fn into_alp(self) -> (Node, Arena<AExpr>, Arena<ALogicalPlan>) {
         let mut expr_arena = Arena::with_capacity(64);
         let mut lp_arena = Arena::with_capacity(32);
-        let root = to_alp(self.logical_plan, &mut expr_arena, &mut lp_arena);
+        let root = to_alp(self.logical_plan, &mut expr_arena, &mut lp_arena).unwrap();
         (root, expr_arena, lp_arena)
     }
 
@@ -225,14 +223,17 @@ impl LazyFrame {
     /// /// Sort DataFrame by 'sepal.width' column
     /// fn example(df: DataFrame) -> LazyFrame {
     ///       df.lazy()
-    ///         .sort("sepal.width", false)
+    ///         .sort("sepal.width", Default::default())
     /// }
     /// ```
-    pub fn sort(self, by_column: &str, reverse: bool) -> Self {
+    pub fn sort(self, by_column: &str, options: SortOptions) -> Self {
+        let reverse = options.descending;
+        let nulls_last = options.nulls_last;
+
         let opt_state = self.get_opt_state();
         let lp = self
             .get_plan_builder()
-            .sort(vec![col(by_column)], vec![reverse])
+            .sort(vec![col(by_column)], vec![reverse], nulls_last)
             .build();
         Self::from_logical_plan(lp, opt_state)
     }
@@ -257,7 +258,10 @@ impl LazyFrame {
             self
         } else {
             let opt_state = self.get_opt_state();
-            let lp = self.get_plan_builder().sort(by_exprs, reverse).build();
+            let lp = self
+                .get_plan_builder()
+                .sort(by_exprs, reverse, false)
+                .build();
             Self::from_logical_plan(lp, opt_state)
         }
     }
@@ -279,30 +283,67 @@ impl LazyFrame {
         self.select_local(vec![col("*").reverse()])
     }
 
-    /// Rename a column in the DataFrame
-    #[deprecated(note = "use rename")]
-    pub fn with_column_renamed(self, existing_name: &str, new_name: &str) -> Self {
-        self.rename([existing_name], [new_name])
-    }
+    fn rename_impl_swapping(self, existing: Vec<String>, new: Vec<String>) -> Self {
+        // schema after renaming
+        let mut new_schema = (&*self.schema()).clone();
 
-    /// Rename columns in the DataFrame.
-    pub fn rename<I, J, T, S>(self, existing: I, new: J) -> Self
-    where
-        I: IntoIterator<Item = T> + Clone,
-        J: IntoIterator<Item = S>,
-        T: AsRef<str>,
-        S: AsRef<str>,
-    {
-        let existing: Vec<String> = existing
-            .into_iter()
-            .map(|name| name.as_ref().to_string())
-            .collect();
+        for (old, new) in existing.iter().zip(new.iter()) {
+            new_schema.rename(old, new.to_string()).unwrap();
+        }
+
+        let prefix = "__POLARS_TEMP_";
 
         let new: Vec<String> = new
-            .into_iter()
-            .map(|name| name.as_ref().to_string())
+            .iter()
+            .map(|name| format!("{}{}", prefix, name))
             .collect();
 
+        self.with_columns(
+            existing
+                .iter()
+                .zip(&new)
+                .map(|(old, new)| col(old).alias(new))
+                .collect::<Vec<_>>(),
+        )
+        .map(
+            move |mut df: DataFrame| {
+                let mut cols = std::mem::take(df.get_columns_mut());
+                // we must find the indices before we start swapping,
+                // because swapping may influence the positions we find if columns are swapped for instance.
+                // e.g. a -> b
+                //      b -> a
+                #[allow(clippy::needless_collect)]
+                let existing_idx = existing
+                    .iter()
+                    .map(|name| cols.iter().position(|s| s.name() == name.as_str()).unwrap())
+                    .collect::<Vec<_>>();
+                let new_idx = new
+                    .iter()
+                    .map(|name| cols.iter().position(|s| s.name() == name.as_str()).unwrap())
+                    .collect::<Vec<_>>();
+
+                for (existing_i, new_i) in existing_idx.into_iter().zip(new_idx) {
+                    cols.swap(existing_i, new_i);
+                    let s = &mut cols[existing_i];
+                    let name = &s.name()[prefix.len()..].to_string();
+                    s.rename(name);
+                }
+                cols.truncate(cols.len() - existing.len());
+                DataFrame::new(cols)
+            },
+            None,
+            Some(new_schema),
+            Some("RENAME_SWAPPING"),
+        )
+    }
+
+    fn rename_imp(self, existing: Vec<String>, new: Vec<String>) -> Self {
+        let mut schema = (*self.schema()).clone();
+
+        for (old, new) in existing.iter().zip(&new) {
+            let _ = schema.rename(old, new.clone());
+        }
+
         self.with_columns(
             existing
                 .iter()
@@ -325,10 +366,40 @@ impl LazyFrame {
                 Ok(df)
             },
             None,
-            None,
+            Some(schema),
+            Some("RENAME"),
         )
     }
 
+    /// Rename columns in the DataFrame.
+    pub fn rename<I, J, T, S>(self, existing: I, new: J) -> Self
+    where
+        I: IntoIterator<Item = T>,
+        J: IntoIterator<Item = S>,
+        T: AsRef<str>,
+        S: AsRef<str>,
+    {
+        // We dispatch to 2 implementations.
+        // 1 is swapping eg. rename a -> b and b -> a
+        // 2 is non-swapping eg. rename a -> new_name
+        // the latter allows predicate pushdown.
+        let existing = existing
+            .into_iter()
+            .map(|a| a.as_ref().to_string())
+            .collect::<Vec<_>>();
+        let new = new
+            .into_iter()
+            .map(|a| a.as_ref().to_string())
+            .collect::<Vec<_>>();
+        let schema = &*self.schema();
+        // a column gets swapped
+        if new.iter().any(|name| schema.get(name).is_some()) {
+            self.rename_impl_swapping(existing, new)
+        } else {
+            self.rename_imp(existing, new)
+        }
+    }
+
     /// Removes columns from the DataFrame.
     /// Note that its better to only select the columns you need
     /// and let the projection pushdown optimize away the unneeded columns.
@@ -419,13 +490,24 @@ impl LazyFrame {
         let logical_plan = self.get_plan_builder().build();
 
         // gradually fill the rules passed to the optimizer
+        let opt = StackOptimizer {};
         let mut rules: Vec<Box<dyn OptimizationRule>> = Vec::with_capacity(8);
 
+        if simplify_expr {
+            rules.push(Box::new(SimplifyExprRule {}));
+            rules.push(Box::new(SimplifyBooleanRule {}));
+        }
+
         // during debug we check if the optimizations have not modified the final schema
         #[cfg(debug_assertions)]
         let prev_schema = logical_plan.schema().clone();
 
-        let mut lp_top = to_alp(logical_plan, expr_arena, lp_arena);
+        let mut lp_top = to_alp(logical_plan, expr_arena, lp_arena)?;
+
+        // simplify expression is valuable for projection and predicate pushdown optimizers, so we
+        // run that first
+        // this optimization will run twice because optimizer may create dumb expressions
+        lp_top = opt.optimize_loop(&mut rules, expr_arena, lp_arena, lp_top);
 
         if projection_pushdown {
             let projection_pushdown_opt = ProjectionPushDown {};
@@ -459,10 +541,6 @@ impl LazyFrame {
             rules.push(Box::new(TypeCoercionRule {}))
         }
 
-        if simplify_expr {
-            rules.push(Box::new(SimplifyExprRule {}));
-            rules.push(Box::new(SimplifyBooleanRule {}));
-        }
         if aggregate_pushdown {
             rules.push(Box::new(AggregatePushdown::new()))
         }
@@ -481,7 +559,6 @@ impl LazyFrame {
         rules.push(Box::new(FastProjection {}));
         rules.push(Box::new(ReplaceDropNulls {}));
 
-        let opt = StackOptimizer {};
         lp_top = opt.optimize_loop(&mut rules, expr_arena, lp_arena, lp_top);
 
         // during debug we check if the optimizations have not modified the final schema
@@ -489,17 +566,11 @@ impl LazyFrame {
         {
             // only check by names because we may supercast types.
             assert_eq!(
-                prev_schema
-                    .fields()
-                    .iter()
-                    .map(|f| f.name())
-                    .collect::<Vec<_>>(),
+                prev_schema.iter_names().collect::<Vec<_>>(),
                 lp_arena
                     .get(lp_top)
                     .schema(lp_arena)
-                    .fields()
-                    .iter()
-                    .map(|f| f.name())
+                    .iter_names()
                     .collect::<Vec<_>>()
             );
         };
@@ -627,7 +698,6 @@ impl LazyFrame {
     ///            col("rain").sum(),
     ///            col("rain").quantile(0.5, QuantileInterpolOptions::Nearest).alias("median_rain"),
     ///        ])
-    ///        .sort("date", false)
     /// }
     /// ```
     pub fn groupby<E: AsRef<[Expr]>>(self, by: E) -> LazyGroupBy {
@@ -638,6 +708,19 @@ impl LazyFrame {
             keys: by.as_ref().to_vec(),
             maintain_order: false,
             dynamic_options: None,
+            rolling_options: None,
+        }
+    }
+
+    pub fn groupby_rolling(self, options: RollingGroupOptions) -> LazyGroupBy {
+        let opt_state = self.get_opt_state();
+        LazyGroupBy {
+            logical_plan: self.logical_plan,
+            opt_state,
+            keys: vec![],
+            maintain_order: true,
+            dynamic_options: None,
+            rolling_options: Some(options),
         }
     }
 
@@ -653,6 +736,7 @@ impl LazyFrame {
             keys: by.as_ref().to_vec(),
             maintain_order: true,
             dynamic_options: Some(options),
+            rolling_options: None,
         }
     }
 
@@ -665,6 +749,7 @@ impl LazyFrame {
             keys: by.as_ref().to_vec(),
             maintain_order: true,
             dynamic_options: None,
+            rolling_options: None,
         }
     }
 
@@ -849,13 +934,35 @@ impl LazyFrame {
         Self::from_logical_plan(lp, opt_state)
     }
 
-    /// Drop duplicate rows. [See eager](polars_core::prelude::DataFrame::drop_duplicates).
-    pub fn drop_duplicates(self, maintain_order: bool, subset: Option<Vec<String>>) -> LazyFrame {
+    /// Keep unique rows and maintain order
+    pub fn distinct_stable(
+        self,
+        subset: Option<Vec<String>>,
+        keep_strategy: DistinctKeepStrategy,
+    ) -> LazyFrame {
         let opt_state = self.get_opt_state();
-        let lp = self
-            .get_plan_builder()
-            .drop_duplicates(maintain_order, subset)
-            .build();
+        let options = DistinctOptions {
+            subset: subset.map(Arc::new),
+            maintain_order: true,
+            keep_strategy,
+        };
+        let lp = self.get_plan_builder().distinct(options).build();
+        Self::from_logical_plan(lp, opt_state)
+    }
+
+    /// Keep unique rows, do not maintain order
+    pub fn distinct(
+        self,
+        subset: Option<Vec<String>>,
+        keep_strategy: DistinctKeepStrategy,
+    ) -> LazyFrame {
+        let opt_state = self.get_opt_state();
+        let options = DistinctOptions {
+            subset: subset.map(Arc::new),
+            maintain_order: false,
+            keep_strategy,
+        };
+        let lp = self.get_plan_builder().distinct(options).build();
         Self::from_logical_plan(lp, opt_state)
     }
 
@@ -890,7 +997,7 @@ impl LazyFrame {
         self.slice(-1, 1)
     }
 
-    /// Get the n last rows
+    /// Get the last `n` rows
     pub fn tail(self, n: u32) -> LazyFrame {
         let neg_tail = -(n as i64);
         self.slice(neg_tail, n)
@@ -924,9 +1031,10 @@ impl LazyFrame {
         function: F,
         optimizations: Option<AllowedOptimizations>,
         schema: Option<Schema>,
+        name: Option<&'static str>,
     ) -> LazyFrame
     where
-        F: DataFrameUdf + 'static,
+        F: 'static + Fn(DataFrame) -> Result<DataFrame> + Send + Sync,
     {
         let opt_state = self.get_opt_state();
         let lp = self
@@ -935,25 +1043,99 @@ impl LazyFrame {
                 function,
                 optimizations.unwrap_or_default(),
                 schema.map(Arc::new),
+                name.unwrap_or("ANONYMOUS UDF"),
             )
             .build();
         Self::from_logical_plan(lp, opt_state)
     }
 
     /// Add a new column at index 0 that counts the rows.
-    pub fn with_row_count(self, name: &str) -> LazyFrame {
+    ///
+    /// # Warning
+    /// This can have a negative effect on query performance.
+    /// This may for instance block predicate pushdown optimization.
+    pub fn with_row_count(mut self, name: &str, offset: Option<u32>) -> LazyFrame {
+        match &mut self.logical_plan {
+            // Do the row count at scan
+            #[cfg(feature = "csv-file")]
+            LogicalPlan::CsvScan { options, .. } => {
+                options.row_count = Some(RowCount {
+                    name: name.to_string(),
+                    offset: offset.unwrap_or(0),
+                });
+                self
+            }
+            #[cfg(feature = "ipc")]
+            LogicalPlan::IpcScan { options, .. } => {
+                options.row_count = Some(RowCount {
+                    name: name.to_string(),
+                    offset: offset.unwrap_or(0),
+                });
+                self
+            }
+            #[cfg(feature = "parquet")]
+            LogicalPlan::ParquetScan { options, .. } => {
+                options.row_count = Some(RowCount {
+                    name: name.to_string(),
+                    offset: offset.unwrap_or(0),
+                });
+                self
+            }
+            _ => {
+                let new_schema = self
+                    .schema()
+                    .insert_index(0, name.to_string(), IDX_DTYPE)
+                    .unwrap();
+                let name = name.to_owned();
+
+                let opt = AllowedOptimizations {
+                    slice_pushdown: false,
+                    predicate_pushdown: false,
+                    ..Default::default()
+                };
+                self.map(
+                    move |df: DataFrame| df.with_row_count(&name, offset),
+                    Some(opt),
+                    Some(new_schema),
+                    Some("WITH ROW COUNT"),
+                )
+            }
+        }
+    }
+
+    /// Unnest the given `Struct` columns. This means that the fields of the `Struct` type will be
+    /// inserted as columns.
+    #[cfg(feature = "dtype-struct")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "dtype-struct")))]
+    pub fn unnest<I: IntoVec<String>>(self, cols: I) -> Self {
+        let cols = cols.into_vec();
+        self.unnest_impl(cols.into_iter().collect())
+    }
+
+    #[cfg(feature = "dtype-struct")]
+    fn unnest_impl(self, cols: PlHashSet<String>) -> Self {
         let schema = self.schema();
 
-        let mut fields = schema.fields().clone();
-        fields.insert(0, Field::new(name, DataType::UInt32));
-        let new_schema = Schema::new(fields);
-
-        let name = name.to_owned();
-
+        let mut new_schema = Schema::with_capacity(schema.len() * 2);
+        for (name, dtype) in schema.iter() {
+            if cols.contains(name) {
+                if let DataType::Struct(flds) = dtype {
+                    for fld in flds {
+                        new_schema.with_column(fld.name().clone(), fld.data_type().clone())
+                    }
+                } else {
+                    // todo: return lazy error here.
+                    panic!("expected struct dtype")
+                }
+            } else {
+                new_schema.with_column(name.clone(), dtype.clone())
+            }
+        }
         self.map(
-            move |df: DataFrame| df.with_row_count(&name),
+            move |df| df.unnest(&cols),
             Some(AllowedOptimizations::default()),
             Some(new_schema),
+            Some("unnest"),
         )
     }
 }
@@ -965,6 +1147,7 @@ pub struct LazyGroupBy {
     keys: Vec<Expr>,
     maintain_order: bool,
     dynamic_options: Option<DynamicGroupOptions>,
+    rolling_options: Option<RollingGroupOptions>,
 }
 
 impl LazyGroupBy {
@@ -982,13 +1165,12 @@ impl LazyGroupBy {
     ///
     /// fn example(df: DataFrame) -> LazyFrame {
     ///       df.lazy()
-    ///        .groupby([col("date")])
+    ///        .groupby_stable([col("date")])
     ///        .agg([
     ///            col("rain").min(),
     ///            col("rain").sum(),
     ///            col("rain").quantile(0.5, QuantileInterpolOptions::Nearest).alias("median_rain"),
     ///        ])
-    ///        .sort("date", false)
     /// }
     /// ```
     pub fn agg<E: AsRef<[Expr]>>(self, aggs: E) -> LazyFrame {
@@ -999,6 +1181,7 @@ impl LazyGroupBy {
                 None,
                 self.maintain_order,
                 self.dynamic_options,
+                self.rolling_options,
             )
             .build();
         LazyFrame::from_logical_plan(lp, self.opt_state)
@@ -1041,6 +1224,7 @@ impl LazyGroupBy {
                 Some(Arc::new(f)),
                 self.maintain_order,
                 None,
+                None,
             )
             .build();
         LazyFrame::from_logical_plan(lp, self.opt_state)
@@ -1057,8 +1241,6 @@ pub struct JoinBuilder {
     allow_parallel: bool,
     force_parallel: bool,
     suffix: Option<String>,
-    asof_by_left: Vec<String>,
-    asof_by_right: Vec<String>,
 }
 impl JoinBuilder {
     pub fn new(lf: LazyFrame) -> Self {
@@ -1071,8 +1253,6 @@ impl JoinBuilder {
             allow_parallel: true,
             force_parallel: false,
             suffix: None,
-            asof_by_left: vec![],
-            asof_by_right: vec![],
         }
     }
 
@@ -1118,13 +1298,6 @@ impl JoinBuilder {
         self
     }
 
-    /// Set the `by` subgrouper of an asof join.
-    pub fn asof_by(mut self, left_by: Vec<String>, right_by: Vec<String>) -> Self {
-        self.asof_by_left = left_by;
-        self.asof_by_right = right_by;
-        self
-    }
-
     /// Finish builder
     pub fn finish(self) -> LazyFrame {
         let opt_state = self.lf.opt_state;
@@ -1146,8 +1319,6 @@ impl JoinBuilder {
                     force_parallel: self.force_parallel,
                     how: self.how,
                     suffix,
-                    asof_by_left: self.asof_by_left,
-                    asof_by_right: self.asof_by_right,
                 },
             )
             .build();
