@@ -4,7 +4,7 @@ macro_rules! push_expr {
     ($current_expr:expr, $push:ident, $iter:ident) => {{
         use Expr::*;
         match $current_expr {
-            Column(_) | Literal(_) | Wildcard => {}
+            Nth(_) | Column(_) | Literal(_) | Wildcard | Columns(_) | DtypeColumn(_) | Count => {}
             Alias(e, _) => $push(e),
             Not(e) => $push(e),
             BinaryExpr { left, op: _, right } => {
@@ -25,7 +25,9 @@ macro_rules! push_expr {
             }
             SortBy { expr, by, .. } => {
                 $push(expr);
-                $push(by)
+                for e in by {
+                    $push(e)
+                }
             }
             Agg(agg_e) => {
                 use AggExpr::*;
@@ -65,6 +67,7 @@ macro_rules! push_expr {
                 function,
                 partition_by,
                 order_by,
+                ..
             } => {
                 $push(function);
                 for e in partition_by {
@@ -74,15 +77,18 @@ macro_rules! push_expr {
                     $push(e);
                 }
             }
-            Slice { input, .. } => $push(input),
-            BinaryFunction {
-                input_a, input_b, ..
+            Slice {
+                input,
+                offset,
+                length,
             } => {
-                $push(input_a);
-                $push(input_b)
+                $push(input);
+                $push(offset);
+                $push(length);
             }
             Exclude(e, _) => $push(e),
             KeepName(e) => $push(e),
+            RenameAlias { expr, .. } => $push(expr),
         }
     }};
 }
@@ -157,7 +163,7 @@ impl AExpr {
         use AExpr::*;
 
         match self {
-            Column(_) | Literal(_) | Wildcard => {}
+            Nth(_) | Column(_) | Literal(_) | Wildcard | Count => {}
             Alias(e, _) => push(e),
             Not(e) => push(e),
             BinaryExpr { left, op: _, right } => {
@@ -174,7 +180,9 @@ impl AExpr {
             }
             SortBy { expr, by, .. } => {
                 push(expr);
-                push(by);
+                for node in by {
+                    push(node)
+                }
             }
             Filter { input, by } => {
                 push(input);
@@ -208,7 +216,7 @@ impl AExpr {
                 push(falsy);
                 push(predicate)
             }
-            Function { input, .. } => input.iter().for_each(|e| push(e)),
+            Function { input, .. } => input.iter().for_each(push),
             Shift { input, .. } => push(input),
             Reverse(e) => push(e),
             Duplicated(e) => push(e),
@@ -218,6 +226,7 @@ impl AExpr {
                 function,
                 partition_by,
                 order_by,
+                options: _,
             } => {
                 push(function);
                 for e in partition_by {
@@ -227,12 +236,14 @@ impl AExpr {
                     push(e);
                 }
             }
-            Slice { input, .. } => push(input),
-            BinaryFunction {
-                input_a, input_b, ..
+            Slice {
+                input,
+                offset,
+                length,
             } => {
-                push(input_a);
-                push(input_b)
+                push(input);
+                push(offset);
+                push(length);
             }
         }
     }
@@ -316,9 +327,9 @@ mod test {
 
         let (root, lp_arena, _expr_arena) = df
             .lazy()
-            .sort("a", false)
-            .groupby(vec![col("a")])
-            .agg(vec![col("a").first()])
+            .sort("a", Default::default())
+            .groupby([col("a")])
+            .agg([col("a").first()])
             .logical_plan
             .into_alp();
 
