@@ -4,11 +4,12 @@ use polars_core::{datatypes::PlHashMap, prelude::*};
 pub(crate) mod aggregate_pushdown;
 #[cfg(any(feature = "parquet", feature = "csv-file"))]
 pub(crate) mod aggregate_scan_projections;
-#[cfg(feature = "private")]
-pub(crate) mod join_pruning;
+pub(crate) mod drop_nulls;
+pub(crate) mod fast_projection;
 pub(crate) mod predicate_pushdown;
 pub(crate) mod projection_pushdown;
 pub(crate) mod simplify_expr;
+pub(crate) mod slice_pushdown;
 pub(crate) mod stack_opt;
 pub(crate) mod type_coercion;
 
