@@ -57,7 +57,7 @@ impl PredicatePushDown {
             // we should not pass these projections
             if exprs
                 .iter()
-                .any(|e_n| is_pushdown_boundary(*e_n, expr_arena))
+                .any(|e_n| other_column_is_pushdown_boundary(*e_n, expr_arena))
             {
                 return self.no_pushdown_restart_opt(lp, acc_predicates, lp_arena, expr_arena);
             }
@@ -72,7 +72,7 @@ impl PredicatePushDown {
             let alp = self.push_down(alp, acc_predicates, lp_arena, expr_arena)?;
             lp_arena.replace(input, alp);
 
-            let lp = lp.from_exprs_and_input(projections, inputs);
+            let lp = lp.with_exprs_and_input(projections, inputs);
             Ok(self.optional_apply_predicate(lp, local_predicates, lp_arena, expr_arena))
         } else {
             let mut local_predicates = Vec::with_capacity(acc_predicates.len());
@@ -81,17 +81,16 @@ impl PredicatePushDown {
             let new_inputs = inputs
                 .iter()
                 .map(|&node| {
-                    // first we check if we are able to push down the predicate pass this node
+                    // first we check if we are able to push down the predicate passed this node
                     // it could be that this node just added the column where we base the predicate on
                     let input_schema = lp_arena.get(node).schema(lp_arena);
                     let mut pushdown_predicates = optimizer::init_hashmap();
-                    for &predicate in acc_predicates.values() {
+                    for (name, &predicate) in acc_predicates.iter() {
                         // we can pushdown the predicate
                         if check_input_node(predicate, input_schema, expr_arena) {
-                            let name = get_insertion_name(expr_arena, predicate, input_schema);
                             insert_and_combine_predicate(
                                 &mut pushdown_predicates,
-                                name,
+                                name.clone(),
                                 predicate,
                                 expr_arena,
                             )
@@ -109,7 +108,7 @@ impl PredicatePushDown {
                 })
                 .collect::<Result<Vec<_>>>()?;
 
-            let lp = lp.from_exprs_and_input(exprs, new_inputs);
+            let lp = lp.with_exprs_and_input(exprs, new_inputs);
             Ok(self.optional_apply_predicate(lp, local_predicates, lp_arena, expr_arena))
         }
     }
@@ -134,7 +133,7 @@ impl PredicatePushDown {
                 Ok(node)
             })
             .collect::<Result<Vec<_>>>()?;
-        let lp = lp.from_exprs_and_input(exprs, new_inputs);
+        let lp = lp.with_exprs_and_input(exprs, new_inputs);
 
         // all predicates are done locally
         let local_predicates = acc_predicates.values().copied().collect::<Vec<_>>();
@@ -301,8 +300,7 @@ impl PredicatePushDown {
             }
             Distinct {
                 input,
-                subset,
-                maintain_order,
+                options
             } => {
                 // currently the distinct operation only keeps the first occurrences.
                 // this may have influence on the pushed down predicates. If the pushed down predicates
@@ -326,8 +324,7 @@ impl PredicatePushDown {
                 self.pushdown_and_assign(input, acc_predicates, lp_arena, expr_arena)?;
                 let lp = Distinct {
                     input,
-                    maintain_order,
-                    subset,
+                    options
                 };
                 Ok(self.optional_apply_predicate(lp, local_predicates, lp_arena, expr_arena))
             }
@@ -350,7 +347,12 @@ impl PredicatePushDown {
                     // unique and duplicated can be caused by joins
                     let matches =
                         |e: &AExpr| matches!(e, AExpr::IsUnique(_) | AExpr::Duplicated(_));
-                    if has_aexpr(predicate, expr_arena, matches) {
+
+                    let checks_nulls =
+                        |e: &AExpr| matches!(e, AExpr::IsNull(_) | AExpr::IsNotNull(_));
+                    if has_aexpr(predicate, expr_arena, matches)
+                        // join might create null values.
+                        || has_aexpr(predicate, expr_arena, checks_nulls) && matches!(&options.how, JoinType::Left | JoinType::Outer | JoinType::Cross){
                         local_predicates.push(predicate);
                         continue;
                     }
@@ -379,7 +381,7 @@ impl PredicatePushDown {
                         );
                         filter_right = true;
                     }
-                    match (filter_left, filter_right, options.how) {
+                    match (filter_left, filter_right, &options.how) {
                         // if not pushed down on of the tables we have to do it locally.
                         (false, false, _) |
                         // if left join and predicate only available in right table,
@@ -421,12 +423,14 @@ impl PredicatePushDown {
 
             lp @ Udf { .. } => {
                 if let ALogicalPlan::Udf {
-                    predicate_pd: true, ..
+                    options: LogicalPlanUdfOptions {
+                        predicate_pd: true, ..
+                    }, ..
                 } = lp
                 {
                     self.pushdown_and_continue(lp, acc_predicates, lp_arena, expr_arena, false)
                 } else {
-                    Ok(lp)
+                    self.no_pushdown_restart_opt(lp, acc_predicates, lp_arena, expr_arena)
                 }
             }
             // Pushed down passed these nodes
@@ -475,7 +479,7 @@ mod test {
             &mut expr_arena,
         );
         let root = *acc_predicates.get("foo").unwrap();
-        let expr = node_to_exp(root, &expr_arena);
+        let expr = node_to_expr(root, &expr_arena);
         assert_eq!(
             format!("{:?}", &expr),
             format!("{:?}", predicate_expr.and(lit(true)))
