@@ -1,5 +1,7 @@
 //! this contains code used for rewriting projections, expanding wildcards, regex selection etc.
 use super::*;
+use crate::utils::has_nth;
+use polars_arrow::index::IndexToUsize;
 
 /// This replace the wilcard Expr with a Column Expr. It also removes the Exclude Expr from the
 /// expression chain.
@@ -20,11 +22,11 @@ pub(super) fn replace_wildcard_with_column(mut expr: Expr, column_name: Arc<str>
     expr
 }
 
-fn rewrite_keep_name_and_sufprefix(expr: Expr) -> Expr {
+fn rewrite_special_aliases(expr: Expr) -> Expr {
     // the blocks are added by cargo fmt
     #[allow(clippy::blocks_in_if_conditions)]
     if has_expr(&expr, |e| {
-        matches!(e, Expr::KeepName(_) | Expr::SufPreFix { .. })
+        matches!(e, Expr::KeepName(_) | Expr::RenameAlias { .. })
     }) {
         match expr {
             Expr::KeepName(expr) => {
@@ -34,18 +36,9 @@ fn rewrite_keep_name_and_sufprefix(expr: Expr) -> Expr {
                     .expect("expected root column to keep expression name");
                 Expr::Alias(expr, name.clone())
             }
-            Expr::SufPreFix {
-                is_suffix,
-                value,
-                expr,
-            } => {
+            Expr::RenameAlias { expr, function } => {
                 let name = get_single_root(&expr).unwrap();
-                let name = if is_suffix {
-                    format!("{}{}", name, value)
-                } else {
-                    format!("{}{}", value, name)
-                };
-
+                let name = function.call(&name);
                 Expr::Alias(expr, Arc::from(name))
             }
             _ => panic!("`keep_name`, `suffix`, `prefix` should be last expression"),
@@ -59,24 +52,41 @@ fn rewrite_keep_name_and_sufprefix(expr: Expr) -> Expr {
 /// with the exclusion of the `names` in the exclude expression.
 /// The resulting expressions are written to result.
 fn replace_wilcard(expr: &Expr, result: &mut Vec<Expr>, exclude: &[Arc<str>], schema: &Schema) {
-    for field in schema.fields() {
-        let name = field.name();
+    for name in schema.iter_names() {
         if !exclude.iter().any(|exluded| &**exluded == name) {
             let new_expr = replace_wildcard_with_column(expr.clone(), Arc::from(name.as_str()));
-            let new_expr = rewrite_keep_name_and_sufprefix(new_expr);
+            let new_expr = rewrite_special_aliases(new_expr);
             result.push(new_expr)
         }
     }
 }
 
+fn replace_nth(expr: &mut Expr, schema: &Schema) {
+    expr.mutate().apply(|e| match e {
+        Expr::Nth(i) => {
+            match i.negative_to_usize(schema.len()) {
+                None => {
+                    let name = if *i == 0 { "first" } else { "last" };
+                    *e = Expr::Column(Arc::from(name));
+                }
+                Some(idx) => {
+                    let (name, _dtype) = schema.get_index(idx).unwrap();
+                    *e = Expr::Column(Arc::from(&**name))
+                }
+            }
+            true
+        }
+        _ => true,
+    })
+}
+
 #[cfg(feature = "regex")]
 /// This function takes an expression containing a regex in `col("..")` and expands the columns
 /// that are selected by that regex in `result`.
 fn expand_regex(expr: &Expr, result: &mut Vec<Expr>, schema: &Schema, pattern: &str) {
     let re = regex::Regex::new(pattern)
         .unwrap_or_else(|_| panic!("invalid regular expression in column: {}", pattern));
-    for field in schema.fields() {
-        let name = field.name();
+    for name in schema.iter_names() {
         if re.is_match(name) {
             let mut new_expr = expr.clone();
 
@@ -88,7 +98,7 @@ fn expand_regex(expr: &Expr, result: &mut Vec<Expr>, schema: &Schema, pattern: &
                 _ => true,
             });
 
-            let new_expr = rewrite_keep_name_and_sufprefix(new_expr);
+            let new_expr = rewrite_special_aliases(new_expr);
             result.push(new_expr)
         }
     }
@@ -106,11 +116,11 @@ fn replace_regex(expr: &Expr, result: &mut Vec<Expr>, schema: &Schema) {
         if name.starts_with('^') && name.ends_with('$') {
             expand_regex(expr, result, schema, name)
         } else {
-            let expr = rewrite_keep_name_and_sufprefix(expr.clone());
+            let expr = rewrite_special_aliases(expr.clone());
             result.push(expr)
         }
     } else {
-        let expr = rewrite_keep_name_and_sufprefix(expr.clone());
+        let expr = rewrite_special_aliases(expr.clone());
         result.push(expr)
     }
 }
@@ -127,7 +137,7 @@ fn expand_columns(expr: &Expr, result: &mut Vec<Expr>, names: &[String]) {
             true
         });
 
-        let new_expr = rewrite_keep_name_and_sufprefix(new_expr);
+        let new_expr = rewrite_special_aliases(new_expr);
         result.push(new_expr)
     }
 }
@@ -135,7 +145,7 @@ fn expand_columns(expr: &Expr, result: &mut Vec<Expr>, names: &[String]) {
 /// replace `DtypeColumn` with `col("foo")..col("bar")`
 fn expand_dtypes(expr: &Expr, result: &mut Vec<Expr>, schema: &Schema, dtypes: &[DataType]) {
     for dtype in dtypes {
-        for field in schema.fields().iter().filter(|f| f.data_type() == dtype) {
+        for field in schema.iter_fields().filter(|f| f.data_type() == dtype) {
             let name = field.name();
 
             let mut new_expr = expr.clone();
@@ -147,7 +157,7 @@ fn expand_dtypes(expr: &Expr, result: &mut Vec<Expr>, schema: &Schema, dtypes: &
                 true
             });
 
-            let new_expr = rewrite_keep_name_and_sufprefix(new_expr);
+            let new_expr = rewrite_special_aliases(new_expr);
             result.push(new_expr)
         }
     }
@@ -178,7 +188,7 @@ fn prepare_excluded(expr: &Expr, schema: &Schema, keys: &[Expr]) -> Vec<Arc<str>
                             }
                         }
                         Excluded::Dtype(dt) => {
-                            for fld in schema.fields() {
+                            for fld in schema.iter_fields() {
                                 if fld.data_type() == dt {
                                     exclude.push(Arc::from(fld.name().as_ref()))
                                 }
@@ -194,9 +204,9 @@ fn prepare_excluded(expr: &Expr, schema: &Schema, keys: &[Expr]) -> Vec<Arc<str>
                     match to_exclude_single {
                         Excluded::Name(name) => exclude.push(name.clone()),
                         Excluded::Dtype(dt) => {
-                            for fld in schema.fields() {
-                                if matches!(fld.data_type(), dt) {
-                                    exclude.push(Arc::from(fld.name().as_ref()))
+                            for (name, dtype) in schema.iter() {
+                                if matches!(dtype, dt) {
+                                    exclude.push(Arc::from(name.as_str()))
                                 }
                             }
                         }
@@ -228,7 +238,7 @@ fn prepare_excluded(expr: &Expr, schema: &Schema, keys: &[Expr]) -> Vec<Arc<str>
 /// In case of single col(*) -> do nothing, no selection is the same as select all
 /// In other cases replace the wildcard with an expression with all columns
 pub(crate) fn rewrite_projections(exprs: Vec<Expr>, schema: &Schema, keys: &[Expr]) -> Vec<Expr> {
-    let mut result = Vec::with_capacity(exprs.len() + schema.fields().len());
+    let mut result = Vec::with_capacity(exprs.len() + schema.len());
 
     for mut expr in exprs {
         // has multiple column names
@@ -244,6 +254,10 @@ pub(crate) fn rewrite_projections(exprs: Vec<Expr>, schema: &Schema, keys: &[Exp
             continue;
         }
 
+        if has_nth(&expr) {
+            replace_nth(&mut expr, schema);
+        }
+
         if has_wildcard(&expr) {
             // keep track of column excluded from the wildcard
             let exclude = prepare_excluded(&expr, schema, keys);
@@ -253,28 +267,32 @@ pub(crate) fn rewrite_projections(exprs: Vec<Expr>, schema: &Schema, keys: &[Exp
                 |e| matches!(e, Expr::Function { options,  .. } if options.input_wildcard_expansion),
             ) {
                 expr.mutate().apply(|e| {
-                    if let Expr::Function { input, .. } = e {
-                        let mut new_inputs = Vec::with_capacity(input.len());
+                    match e {
+                        Expr::Function { input, options, .. }
+                            if options.input_wildcard_expansion =>
+                        {
+                            let mut new_inputs = Vec::with_capacity(input.len());
 
-                        input.iter_mut().for_each(|e| {
-                            if has_wildcard(e) {
-                                replace_wilcard(e, &mut new_inputs, &exclude, schema)
-                            } else {
-                                #[cfg(feature = "regex")]
-                                {
-                                    replace_regex(e, &mut new_inputs, schema)
-                                }
-                                #[cfg(not(feature = "regex"))]
-                                {
-                                    new_inputs.push(e.clone())
-                                }
-                            }
-                        });
+                            input.iter_mut().for_each(|e| {
+                                if has_wildcard(e) {
+                                    replace_wilcard(e, &mut new_inputs, &exclude, schema)
+                                } else {
+                                    #[cfg(feature = "regex")]
+                                    {
+                                        replace_regex(e, &mut new_inputs, schema)
+                                    }
+                                    #[cfg(not(feature = "regex"))]
+                                    {
+                                        new_inputs.push(e.clone())
+                                    }
+                                };
+                            });
 
-                        *input = new_inputs;
-                        false
-                    } else {
-                        true
+                            *input = new_inputs;
+                            // continue there can be more functions that require expansion
+                            true
+                        }
+                        _ => true,
                     }
                 });
                 result.push(expr);
@@ -289,7 +307,7 @@ pub(crate) fn rewrite_projections(exprs: Vec<Expr>, schema: &Schema, keys: &[Exp
             }
             #[cfg(not(feature = "regex"))]
             {
-                let expr = rewrite_keep_name_and_sufprefix(expr);
+                let expr = rewrite_special_aliases(expr);
                 result.push(expr)
             }
         };
