@@ -1,9 +1,5 @@
 use super::*;
-use crate::logical_plan::CsvParserOptions;
-#[cfg(feature = "ipc")]
-use crate::logical_plan::LpScanOptions;
-#[cfg(feature = "parquet")]
-use crate::logical_plan::ParquetOptions;
+use crate::prelude::*;
 use crate::utils::try_path_to_str;
 use polars_io::aggregations::ScanAggregation;
 use polars_io::csv::CsvEncoding;
@@ -53,7 +49,7 @@ fn prepare_scan_args<'a>(
     let projection: Option<Vec<_>> = with_columns.map(|with_columns| {
         with_columns
             .iter()
-            .map(|name| schema.column_with_name(name).unwrap().0)
+            .map(|name| schema.index_of(name).unwrap())
             .collect()
     });
 
@@ -76,7 +72,7 @@ pub struct IpcExec {
     pub(crate) schema: SchemaRef,
     pub(crate) predicate: Option<Arc<dyn PhysicalExpr>>,
     pub(crate) aggregate: Vec<ScanAggregation>,
-    pub(crate) options: LpScanOptions,
+    pub(crate) options: IpcScanOptions,
 }
 
 #[cfg(feature = "ipc")]
@@ -96,6 +92,7 @@ impl Executor for IpcExec {
         );
         let df = IpcReader::new(file)
             .with_n_rows(n_rows)
+            .with_row_count(std::mem::take(&mut self.options.row_count))
             .finish_with_scan_ops(
                 predicate,
                 aggregate,
@@ -160,6 +157,7 @@ impl Executor for ParquetExec {
         let df = ParquetReader::new(file)
             .with_n_rows(n_rows)
             .read_parallel(self.options.parallel)
+            .with_row_count(std::mem::take(&mut self.options.row_count))
             .finish_with_scan_ops(
                 predicate,
                 aggregate,
@@ -228,13 +226,15 @@ impl Executor for CsvExec {
             .with_n_rows(n_rows)
             .with_columns(with_columns)
             .low_memory(self.options.low_memory)
-            .with_null_values(self.options.null_values.clone())
+            .with_null_values(std::mem::take(&mut self.options.null_values))
             .with_predicate(predicate)
             .with_aggregate(aggregate)
             .with_encoding(CsvEncoding::LossyUtf8)
             .with_comment_char(self.options.comment_char)
             .with_quote_char(self.options.quote_char)
+            .with_encoding(self.options.encoding)
             .with_rechunk(self.options.rechunk)
+            .with_row_count(std::mem::take(&mut self.options.row_count))
             .finish()?;
 
         if self.options.cache {
@@ -264,6 +264,7 @@ impl Executor for DataFrameExec {
         // projection should be before selection as those are free
         // TODO: this is only the case if we don't create new columns
         if let Some(projection) = &self.projection {
+            state.may_set_schema(&df, projection.len());
             df = evaluate_physical_expressions(&df, projection, state, self.has_windows)?;
         }
 
@@ -274,6 +275,7 @@ impl Executor for DataFrameExec {
             })?;
             df = df.filter(mask)?;
         }
+        state.clear_schema_cache();
 
         if let Some(limit) = set_n_rows(None) {
             Ok(df.head(Some(limit)))
