@@ -1,8 +1,10 @@
 use crate::physical_plan::state::ExecutionState;
 use crate::physical_plan::PhysicalAggregation;
 use crate::prelude::*;
-use polars_core::frame::groupby::GroupTuples;
+use polars_arrow::utils::CustomIterTools;
+use polars_core::frame::groupby::GroupsProxy;
 use polars_core::prelude::*;
+use polars_core::POOL;
 use rayon::prelude::*;
 use std::sync::Arc;
 
@@ -19,13 +21,31 @@ impl ApplyExpr {
     fn prepare_multiple_inputs<'a>(
         &self,
         df: &DataFrame,
-        groups: &'a GroupTuples,
+        groups: &'a GroupsProxy,
         state: &ExecutionState,
     ) -> Result<Vec<AggregationContext<'a>>> {
-        self.inputs
-            .par_iter()
-            .map(|e| e.evaluate_on_groups(df, groups, state))
-            .collect()
+        POOL.install(|| {
+            self.inputs
+                .par_iter()
+                .map(|e| e.evaluate_on_groups(df, groups, state))
+                .collect()
+        })
+    }
+
+    fn finish_apply_groups<'a>(
+        &self,
+        mut ac: AggregationContext<'a>,
+        ca: ListChunked,
+        all_unit_len: bool,
+    ) -> AggregationContext<'a> {
+        if all_unit_len && self.auto_explode {
+            ac.with_series(ca.explode().unwrap().into_series(), true);
+        } else {
+            ac.with_series(ca.into_series(), true);
+        }
+        ac.with_all_unit_len(all_unit_len);
+        ac.with_update_groups(UpdateGroups::WithSeriesLen);
+        ac
     }
 }
 
@@ -51,7 +71,7 @@ impl PhysicalExpr for ApplyExpr {
     fn evaluate_on_groups<'a>(
         &self,
         df: &DataFrame,
-        groups: &'a GroupTuples,
+        groups: &'a GroupsProxy,
         state: &ExecutionState,
     ) -> Result<AggregationContext<'a>> {
         if self.inputs.len() == 1 {
@@ -92,15 +112,18 @@ impl PhysicalExpr for ApplyExpr {
                         .collect();
 
                     ca.rename(&name);
-                    ac.with_series(ca.into_series(), true);
-                    ac.with_all_unit_len(all_unit_len);
-                    ac.with_update_groups(UpdateGroups::WithSeriesLen);
+                    let ac = self.finish_apply_groups(ac, ca, all_unit_len);
                     Ok(ac)
                 }
                 ApplyOptions::ApplyFlat => {
-                    let s = self
-                        .function
-                        .call_udf(&mut [ac.flat_naive().into_owned()])?;
+                    let input = ac.flat_naive().into_owned();
+                    let input_len = input.len();
+                    let s = self.function.call_udf(&mut [input])?;
+
+                    if s.len() != input_len {
+                        return Err(PolarsError::ComputeError("A map function may never return a Series of a different length than its input".into()));
+                    }
+
                     if ac.is_aggregated() {
                         ac.with_update_groups(UpdateGroups::WithGroupsLen);
                     }
@@ -108,9 +131,7 @@ impl PhysicalExpr for ApplyExpr {
                     Ok(ac)
                 }
                 ApplyOptions::ApplyList => {
-                    let s = self
-                        .function
-                        .call_udf(&mut [ac.aggregated().into_owned()])?;
+                    let s = self.function.call_udf(&mut [ac.aggregated()])?;
                     ac.with_series(s, true);
                     Ok(ac)
                 }
@@ -154,11 +175,10 @@ impl PhysicalExpr for ApplyExpr {
                                 s
                             })
                         })
-                        .collect();
+                        .collect_trusted();
                     ca.rename(&name);
-                    let mut ac = acs.pop().unwrap();
-                    ac.with_series(ca.into_series(), true);
-                    ac.with_all_unit_len(all_unit_len);
+                    let ac = acs.pop().unwrap();
+                    let ac = self.finish_apply_groups(ac, ca, all_unit_len);
                     Ok(ac)
                 }
                 ApplyOptions::ApplyFlat => {
@@ -169,15 +189,14 @@ impl PhysicalExpr for ApplyExpr {
 
                     let s = self.function.call_udf(&mut s)?;
                     let mut ac = acs.pop().unwrap();
-                    ac.with_update_groups(UpdateGroups::WithGroupsLen);
+                    if ac.is_aggregated() {
+                        ac.with_update_groups(UpdateGroups::WithGroupsLen);
+                    }
                     ac.with_series(s, false);
                     Ok(ac)
                 }
                 ApplyOptions::ApplyList => {
-                    let mut s = acs
-                        .iter_mut()
-                        .map(|ac| ac.aggregated().into_owned())
-                        .collect::<Vec<_>>();
+                    let mut s = acs.iter_mut().map(|ac| ac.aggregated()).collect::<Vec<_>>();
                     let s = self.function.call_udf(&mut s)?;
                     let mut ac = acs.pop().unwrap();
                     ac.with_update_groups(UpdateGroups::WithGroupsLen);
@@ -200,14 +219,11 @@ impl PhysicalAggregation for ApplyExpr {
     fn aggregate(
         &self,
         df: &DataFrame,
-        groups: &GroupTuples,
+        groups: &GroupsProxy,
         state: &ExecutionState,
     ) -> Result<Option<Series>> {
         let mut ac = self.evaluate_on_groups(df, groups, state)?;
-        let mut s = ac.aggregated().into_owned();
-        if ac.is_all_unit_len() && self.auto_explode {
-            s = s.explode().unwrap();
-        }
+        let s = ac.aggregated();
         Ok(Some(s))
     }
 }
