@@ -1,7 +1,7 @@
 use crate::physical_plan::state::ExecutionState;
 use crate::physical_plan::PhysicalAggregation;
 use crate::prelude::*;
-use polars_core::frame::groupby::GroupTuples;
+use polars_core::frame::groupby::GroupsProxy;
 use polars_core::series::unstable::UnstableSeries;
 use polars_core::{prelude::*, POOL};
 use std::convert::TryFrom;
@@ -42,7 +42,6 @@ pub(crate) fn apply_operator(left: &Series, right: &Series, op: Operator) -> Res
         Operator::Minus => Ok(left - right),
         Operator::Multiply => Ok(left * right),
         Operator::Divide => Ok(left / right),
-        #[cfg(feature = "true_div")]
         Operator::TrueDivide => {
             use DataType::*;
             match left.dtype() {
@@ -76,7 +75,7 @@ impl PhysicalExpr for BinaryExpr {
     fn evaluate_on_groups<'a>(
         &self,
         df: &DataFrame,
-        groups: &'a GroupTuples,
+        groups: &'a GroupsProxy,
         state: &ExecutionState,
     ) -> Result<AggregationContext<'a>> {
         let (result_a, result_b) = POOL.install(|| {
@@ -96,21 +95,53 @@ impl PhysicalExpr for BinaryExpr {
             ));
         }
 
-        match (ac_l.agg_state(), ac_r.agg_state()) {
+        match (ac_l.agg_state(), ac_r.agg_state(), self.op) {
+            // Some aggregations must return boolean masks that fit the group. That's why not all literals can take this path.
+            // only literals that are used in arithmetic
+            (
+                AggState::AggregatedFlat(lhs),
+                AggState::Literal(rhs),
+                Operator::Plus
+                | Operator::Minus
+                | Operator::Divide
+                | Operator::Multiply
+                | Operator::Modulus
+                | Operator::TrueDivide,
+            )
+            | (
+                AggState::Literal(lhs),
+                AggState::AggregatedFlat(rhs),
+                Operator::Plus
+                | Operator::Minus
+                | Operator::Divide
+                | Operator::Multiply
+                | Operator::Modulus
+                | Operator::TrueDivide,
+            ) => {
+                let out = apply_operator(lhs, rhs, self.op)?;
+
+                ac_l.with_series(out, true);
+                Ok(ac_l)
+            }
             // One of the two exprs is aggregated with flat aggregation, e.g. `e.min(), e.max(), e.first()`
 
             // if the groups_len == df.len we can just apply all flat.
-            (AggState::AggregatedFlat(s), AggState::NotAggregated(_)) if s.len() != df.height() => {
+            // within an aggregation a `col().first() - lit(0)` must still produce a boolean array of group length,
+            // that's why a literal also takes this branch
+            (AggState::AggregatedFlat(s), AggState::NotAggregated(_) | AggState::Literal(_), _)
+                if s.len() != df.height() =>
+            {
                 // this is a flat series of len eq to group tuples
                 let l = ac_l.aggregated();
                 let l = l.as_ref();
                 let arr_l = &l.chunks()[0];
-                assert_eq!(l.len(), groups.len());
 
                 // we create a dummy Series that is not cloned nor moved
                 // so we can swap the ArrayRef during the hot loop
                 // this prevents a series Arc alloc and a vec alloc per iteration
                 let dummy = Series::try_from(("dummy", vec![arr_l.clone()])).unwrap();
+                // keep logical type info
+                let dummy = dummy.cast(l.dtype()).unwrap();
                 let mut us = UnstableSeries::new(&dummy);
 
                 // this is now a list
@@ -141,10 +172,15 @@ impl PhysicalExpr for BinaryExpr {
                 ca.rename(l.name());
 
                 ac_l.with_series(ca.into_series(), true);
+                ac_l.with_update_groups(UpdateGroups::WithGroupsLen);
                 Ok(ac_l)
             }
             // if the groups_len == df.len we can just apply all flat.
-            (AggState::NotAggregated(_), AggState::AggregatedFlat(s)) if s.len() != df.height() => {
+            (
+                AggState::Literal(_) | AggState::AggregatedList(_) | AggState::NotAggregated(_),
+                AggState::AggregatedFlat(s),
+                _,
+            ) if s.len() != df.height() => {
                 // this is now a list
                 let l = ac_l.aggregated();
                 let l = l.list().unwrap();
@@ -159,6 +195,8 @@ impl PhysicalExpr for BinaryExpr {
                 // so we can swap the ArrayRef during the hot loop
                 // this prevents a series Arc alloc and a vec alloc per iteration
                 let dummy = Series::try_from(("dummy", vec![arr_r.clone()])).unwrap();
+                // keep logical type info
+                let dummy = dummy.cast(r.dtype()).unwrap();
                 let mut us = UnstableSeries::new(&dummy);
 
                 let mut ca: ListChunked = l
@@ -183,9 +221,39 @@ impl PhysicalExpr for BinaryExpr {
                 ca.rename(l.name());
 
                 ac_l.with_series(ca.into_series(), true);
+                ac_l.with_update_groups(UpdateGroups::WithGroupsLen);
                 Ok(ac_l)
             }
-            // Both are or a flat series or aggregated into a list
+            (AggState::AggregatedList(_), AggState::NotAggregated(_) | AggState::Literal(_), _)
+            | (AggState::NotAggregated(_) | AggState::Literal(_), AggState::AggregatedList(_), _) =>
+            {
+                ac_l.sort_by_groups();
+                ac_r.sort_by_groups();
+
+                let out = apply_operator(
+                    ac_l.flat_naive().as_ref(),
+                    ac_r.flat_naive().as_ref(),
+                    self.op,
+                )?;
+
+                // we flattened the series, so that sorts by group
+                ac_l.with_update_groups(UpdateGroups::WithGroupsLen);
+                ac_l.with_series(out, false);
+                Ok(ac_l)
+            }
+            // flatten the Series and apply the operators
+            (AggState::AggregatedList(_), AggState::AggregatedList(_), _) => {
+                let out = apply_operator(
+                    ac_l.flat_naive().as_ref(),
+                    ac_r.flat_naive().as_ref(),
+                    self.op,
+                )?;
+
+                ac_l.combine_groups(ac_r).with_series(out, false);
+                ac_l.with_update_groups(UpdateGroups::WithGroupsLen);
+                Ok(ac_l)
+            }
+            // Both are or a flat series
             // so we can flatten the Series and apply the operators
             _ => {
                 let out = apply_operator(
@@ -193,14 +261,15 @@ impl PhysicalExpr for BinaryExpr {
                     ac_r.flat_naive().as_ref(),
                     self.op,
                 )?;
+
                 ac_l.combine_groups(ac_r).with_series(out, false);
                 Ok(ac_l)
             }
         }
     }
 
-    fn to_field(&self, _input_schema: &Schema) -> Result<Field> {
-        todo!()
+    fn to_field(&self, input_schema: &Schema) -> Result<Field> {
+        self.expr.to_field(input_schema, Context::Default)
     }
 
     fn as_agg_expr(&self) -> Result<&dyn PhysicalAggregation> {
@@ -216,7 +285,7 @@ impl PhysicalAggregation for BinaryExpr {
     fn aggregate(
         &self,
         df: &DataFrame,
-        groups: &GroupTuples,
+        groups: &GroupsProxy,
         state: &ExecutionState,
     ) -> Result<Option<Series>> {
         match (self.left.as_agg_expr(), self.right.as_agg_expr()) {
@@ -253,25 +322,37 @@ mod stats {
 
     fn apply_operator_stats_rhs_lit(min_max: &Series, literal: &Series, op: Operator) -> bool {
         match op {
+            // col > lit
+            // e.g.
+            // [min,
+            // max] > 0
+            //
+            // [-1,
+            // 2] > 0
+            //
+            // [false, true] -> true -> read
             Operator::Gt => {
                 // literal is bigger than max value
                 // selection needs all rows
-                ChunkCompare::<&Series>::gt(min_max, literal).all()
+                ChunkCompare::<&Series>::gt(min_max, literal).any()
             }
+            // col >= lit
             Operator::GtEq => {
                 // literal is bigger than max value
                 // selection needs all rows
-                ChunkCompare::<&Series>::gt_eq(min_max, literal).all()
+                ChunkCompare::<&Series>::gt_eq(min_max, literal).any()
             }
+            // col < lit
             Operator::Lt => {
                 // literal is smaller than min value
                 // selection needs all rows
-                ChunkCompare::<&Series>::lt(min_max, literal).all()
+                ChunkCompare::<&Series>::lt(min_max, literal).any()
             }
+            // col <= lit
             Operator::LtEq => {
                 // literal is smaller than min value
                 // selection needs all rows
-                ChunkCompare::<&Series>::lt_eq(min_max, literal).all()
+                ChunkCompare::<&Series>::lt_eq(min_max, literal).any()
             }
             // default: read the file
             _ => true,
@@ -283,22 +364,22 @@ mod stats {
             Operator::Gt => {
                 // literal is bigger than max value
                 // selection needs all rows
-                ChunkCompare::<&Series>::gt(literal, min_max).all()
+                ChunkCompare::<&Series>::gt(literal, min_max).any()
             }
             Operator::GtEq => {
                 // literal is bigger than max value
                 // selection needs all rows
-                ChunkCompare::<&Series>::gt_eq(literal, min_max).all()
+                ChunkCompare::<&Series>::gt_eq(literal, min_max).any()
             }
             Operator::Lt => {
                 // literal is smaller than min value
                 // selection needs all rows
-                ChunkCompare::<&Series>::lt(literal, min_max).all()
+                ChunkCompare::<&Series>::lt(literal, min_max).any()
             }
             Operator::LtEq => {
                 // literal is smaller than min value
                 // selection needs all rows
-                ChunkCompare::<&Series>::lt_eq(literal, min_max).all()
+                ChunkCompare::<&Series>::lt_eq(literal, min_max).any()
             }
             // default: read the file
             _ => true,
@@ -366,12 +447,6 @@ mod stats {
                     Operator::Or => Ok(l.should_read(stats)? || r.should_read(stats)?),
                     _ => Ok(true),
                 },
-                // This branch is probably never hit
-                (Some(other), None) | (None, Some(other)) => match self.op {
-                    Operator::And => Ok(self.should_read(stats)? && other.should_read(stats)?),
-                    Operator::Or => Ok(self.should_read(stats)? || other.should_read(stats)?),
-                    _ => Ok(true),
-                },
                 _ => self.impl_should_read(stats),
             }
         }
