@@ -5,19 +5,20 @@ use crate::physical_plan::executors::groupby_rolling::GroupByRollingExec;
 #[cfg(feature = "ipc")]
 use crate::physical_plan::executors::scan::IpcExec;
 use crate::physical_plan::executors::union::UnionExec;
+use crate::prelude::count::CountExpr;
 use crate::prelude::shift::ShiftExpr;
 use crate::prelude::*;
-use crate::utils::{expr_to_root_column_name, has_window_aexpr};
+#[cfg(feature = "object")]
+use crate::utils::expr_to_root_column_name;
+use crate::utils::has_window_aexpr;
 use crate::{
     logical_plan::iterator::ArenaExprIter,
     utils::{aexpr_to_root_names, aexpr_to_root_nodes, agg_source_paths, has_aexpr},
 };
-use ahash::RandomState;
 use polars_core::prelude::*;
 use polars_core::{frame::groupby::GroupByMethod, utils::parallel_op_series};
 #[cfg(any(feature = "parquet", feature = "csv-file", feature = "ipc"))]
 use polars_io::aggregations::ScanAggregation;
-use std::collections::HashSet;
 use std::sync::Arc;
 
 #[cfg(any(feature = "parquet", feature = "csv-file"))]
@@ -196,6 +197,7 @@ impl DefaultPlanner {
                 schema: _schema,
                 ..
             } => {
+                let input_schema = lp_arena.get(input).schema(lp_arena).clone();
                 let has_windows = expr.iter().any(|node| has_window_aexpr(*node, expr_arena));
                 let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 let phys_expr =
@@ -204,6 +206,7 @@ impl DefaultPlanner {
                     input,
                     expr: phys_expr,
                     has_windows,
+                    input_schema,
                     #[cfg(test)]
                     schema: _schema,
                 }))
@@ -211,9 +214,12 @@ impl DefaultPlanner {
             LocalProjection {
                 expr,
                 input,
-                schema: _schema,
+                #[cfg(test)]
+                    schema: _schema,
                 ..
             } => {
+                let input_schema = lp_arena.get(input).schema(lp_arena).clone();
+
                 let has_windows = expr.iter().any(|node| has_window_aexpr(*node, expr_arena));
                 let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 let phys_expr =
@@ -222,6 +228,7 @@ impl DefaultPlanner {
                     input,
                     expr: phys_expr,
                     has_windows,
+                    input_schema,
                     #[cfg(test)]
                     schema: _schema,
                 }))
@@ -258,7 +265,7 @@ impl DefaultPlanner {
             Sort {
                 input,
                 by_column,
-                reverse,
+                args,
             } => {
                 let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 let by_column =
@@ -266,7 +273,7 @@ impl DefaultPlanner {
                 Ok(Box::new(SortExec {
                     input,
                     by_column,
-                    reverse,
+                    args,
                 }))
             }
             Explode { input, columns } => {
@@ -274,32 +281,24 @@ impl DefaultPlanner {
                 Ok(Box::new(ExplodeExec { input, columns }))
             }
             Cache { input } => {
-                let fields = lp_arena.get(input).schema(lp_arena).fields();
+                let schema = lp_arena.get(input).schema(lp_arena);
                 // todo! fix the unique constraint in the schema. Probably in projection pushdown at joins
-                let mut unique =
-                    HashSet::with_capacity_and_hasher(fields.len(), RandomState::default());
+                let mut unique = PlHashSet::with_capacity(schema.len());
                 // assumption of 80 characters per column name
-                let mut key = String::with_capacity(fields.len() * 80);
-                for field in fields {
-                    if unique.insert(field.name()) {
-                        key.push_str(field.name())
+                let mut key = String::with_capacity(schema.len() * 80);
+                for name in schema.iter_names() {
+                    if unique.insert(name) {
+                        key.push_str(name)
                     }
                 }
+                // mutable borrow otherwise
+                drop(unique);
                 let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 Ok(Box::new(CacheExec { key, input }))
             }
-            Distinct {
-                input,
-                maintain_order,
-                subset,
-            } => {
+            Distinct { input, options } => {
                 let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
-                let subset = Arc::try_unwrap(subset).unwrap_or_else(|subset| (*subset).clone());
-                Ok(Box::new(DropDuplicatesExec {
-                    input,
-                    maintain_order,
-                    subset,
-                }))
+                Ok(Box::new(DropDuplicatesExec { input, options }))
             }
             Aggregate {
                 input,
@@ -310,7 +309,6 @@ impl DefaultPlanner {
                 maintain_order,
                 options,
             } => {
-                #[cfg(feature = "object")]
                 let input_schema = lp_arena.get(input).schema(lp_arena).clone();
                 let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
 
@@ -326,6 +324,7 @@ impl DefaultPlanner {
                         keys: phys_keys,
                         aggs: phys_aggs,
                         options,
+                        input_schema,
                     }));
                 }
 
@@ -334,6 +333,7 @@ impl DefaultPlanner {
                         input,
                         aggs: phys_aggs,
                         options,
+                        input_schema,
                     }));
                 }
 
@@ -345,8 +345,7 @@ impl DefaultPlanner {
                 //      1. complex expressions in the groupby itself are also not partitionable
                 //          in this case anything more than col("foo")
                 //      2. a custom function cannot be partitioned
-                //      3. maintain order is likely cheaper in default groupby
-                if keys.len() == 1 && apply.is_none() && !maintain_order {
+                if keys.len() == 1 && apply.is_none() {
                     // complex expressions in the groupby itself are also not partitionable
                     // in this case anything more than col("foo")
                     if (&*expr_arena).iter(keys[0]).count() > 1 {
@@ -361,7 +360,6 @@ impl DefaultPlanner {
                                 AExpr::SortBy { .. }
                                     | AExpr::Filter { .. }
                                     | AExpr::BinaryExpr { .. }
-                                    | AExpr::BinaryFunction { .. }
                                     | AExpr::Function { .. }
                             )
                         };
@@ -372,14 +370,14 @@ impl DefaultPlanner {
                             break;
                         }
 
-                        let agg = node_to_exp(*agg, expr_arena);
+                        let agg = node_to_expr(*agg, expr_arena);
 
                         #[cfg(feature = "object")]
                         {
                             let name = expr_to_root_column_name(&agg).unwrap();
-                            let fld = input_schema.field_with_name(&name).unwrap();
+                            let dtype = input_schema.get(&name).unwrap();
 
-                            if let DataType::Object(_) = fld.data_type() {
+                            if let DataType::Object(_) = dtype {
                                 partitionable = false;
                                 break;
                             }
@@ -411,8 +409,9 @@ impl DefaultPlanner {
                         phys_keys.pop().unwrap(),
                         phys_aggs,
                         aggs.into_iter()
-                            .map(|n| node_to_exp(n, expr_arena))
+                            .map(|n| node_to_expr(n, expr_arena))
                             .collect(),
+                        maintain_order,
                     )))
                 } else {
                     Ok(Box::new(GroupByExec::new(
@@ -421,6 +420,7 @@ impl DefaultPlanner {
                         phys_aggs,
                         apply,
                         maintain_order,
+                        input_schema,
                     )))
                 }
             }
@@ -438,11 +438,9 @@ impl DefaultPlanner {
                     // check if two DataFrames come from a separate source.
                     // If they don't we can parallelize,
                     // Otherwise it is in cache.
-                    let mut sources_left =
-                        HashSet::with_capacity_and_hasher(16, RandomState::default());
+                    let mut sources_left = PlHashSet::with_capacity(16);
                     agg_source_paths(input_left, &mut sources_left, lp_arena);
-                    let mut sources_right =
-                        HashSet::with_capacity_and_hasher(16, RandomState::default());
+                    let mut sources_right = PlHashSet::with_capacity(16);
                     agg_source_paths(input_right, &mut sources_right, lp_arena);
                     sources_left.intersection(&sources_right).next().is_none()
                 } else {
@@ -463,11 +461,10 @@ impl DefaultPlanner {
                     right_on,
                     parallel,
                     options.suffix,
-                    options.asof_by_left,
-                    options.asof_by_right,
                 )))
             }
             HStack { input, exprs, .. } => {
+                let input_schema = lp_arena.get(input).schema(lp_arena).clone();
                 let has_windows = exprs.iter().any(|node| has_window_aexpr(*node, expr_arena));
                 let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 let phys_expr =
@@ -476,6 +473,7 @@ impl DefaultPlanner {
                     input,
                     has_windows,
                     expr: phys_expr,
+                    input_schema,
                 }))
             }
             Udf {
@@ -496,6 +494,7 @@ impl DefaultPlanner {
         use AExpr::*;
 
         match expr_arena.get(expression).clone() {
+            Count => Ok(Arc::new(CountExpr::new())),
             Window {
                 mut function,
                 partition_by,
@@ -517,8 +516,8 @@ impl DefaultPlanner {
                     if has_aexpr(function, expr_arena, |e| matches!(e, AExpr::Literal(_))) {
                         apply_columns.push(Arc::from("literal"))
                     } else {
-                        let e = node_to_exp(function, expr_arena);
-                        return Err(PolarsError::ValueError(
+                        let e = node_to_expr(function, expr_arena);
+                        return Err(PolarsError::ComputeError(
                             format!(
                                 "Cannot apply a window function, did not find a root column. \
                             This is likely due to a syntax error in this expression: {:?}",
@@ -533,7 +532,7 @@ impl DefaultPlanner {
                     function = *expr;
                     out_name = Some(name.clone());
                 };
-                let function = node_to_exp(function, expr_arena);
+                let function = node_to_expr(function, expr_arena);
 
                 Ok(Arc::new(WindowExpr {
                     group_by,
@@ -542,12 +541,12 @@ impl DefaultPlanner {
                     function,
                     phys_function,
                     options,
-                    expr: node_to_exp(expression, expr_arena),
+                    expr: node_to_expr(expression, expr_arena),
                 }))
             }
             Literal(value) => Ok(Arc::new(LiteralExpr::new(
                 value,
-                node_to_exp(expression, expr_arena),
+                node_to_expr(expression, expr_arena),
             ))),
             BinaryExpr { left, op, right } => {
                 let lhs = self.create_physical_expr(left, ctxt, expr_arena)?;
@@ -556,19 +555,19 @@ impl DefaultPlanner {
                     lhs,
                     op,
                     rhs,
-                    node_to_exp(expression, expr_arena),
+                    node_to_expr(expression, expr_arena),
                 )))
             }
             Column(column) => Ok(Arc::new(ColumnExpr::new(
                 column,
-                node_to_exp(expression, expr_arena),
+                node_to_expr(expression, expr_arena),
             ))),
             Sort { expr, options } => {
                 let phys_expr = self.create_physical_expr(expr, ctxt, expr_arena)?;
                 Ok(Arc::new(SortExpr::new(
                     phys_expr,
                     options,
-                    node_to_exp(expression, expr_arena),
+                    node_to_expr(expression, expr_arena),
                 )))
             }
             Take { expr, idx } => {
@@ -577,7 +576,7 @@ impl DefaultPlanner {
                 Ok(Arc::new(TakeExpr {
                     phys_expr,
                     idx: phys_idx,
-                    expr: node_to_exp(expression, expr_arena),
+                    expr: node_to_expr(expression, expr_arena),
                 }))
             }
             SortBy { expr, by, reverse } => {
@@ -587,7 +586,7 @@ impl DefaultPlanner {
                     phys_expr,
                     phys_by,
                     reverse,
-                    node_to_exp(expression, expr_arena),
+                    node_to_expr(expression, expr_arena),
                 )))
             }
             Filter { input, by } => {
@@ -596,14 +595,14 @@ impl DefaultPlanner {
                 Ok(Arc::new(FilterExpr::new(
                     phys_input,
                     phys_by,
-                    node_to_exp(expression, expr_arena),
+                    node_to_expr(expression, expr_arena),
                 )))
             }
             Not(expr) => {
                 let phys_expr = self.create_physical_expr(expr, ctxt, expr_arena)?;
                 Ok(Arc::new(NotExpr::new(
                     phys_expr,
-                    node_to_exp(expression, expr_arena),
+                    node_to_expr(expression, expr_arena),
                 )))
             }
             Alias(expr, name) => {
@@ -611,21 +610,21 @@ impl DefaultPlanner {
                 Ok(Arc::new(AliasExpr::new(
                     phys_expr,
                     name,
-                    node_to_exp(expression, expr_arena),
+                    node_to_expr(expression, expr_arena),
                 )))
             }
             IsNull(expr) => {
                 let phys_expr = self.create_physical_expr(expr, ctxt, expr_arena)?;
                 Ok(Arc::new(IsNullExpr::new(
                     phys_expr,
-                    node_to_exp(expression, expr_arena),
+                    node_to_expr(expression, expr_arena),
                 )))
             }
             IsNotNull(expr) => {
                 let phys_expr = self.create_physical_expr(expr, ctxt, expr_arena)?;
                 Ok(Arc::new(IsNotNullExpr::new(
                     phys_expr,
-                    node_to_exp(expression, expr_arena),
+                    node_to_expr(expression, expr_arena),
                 )))
             }
             Agg(agg) => {
@@ -646,7 +645,7 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -668,7 +667,7 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -690,7 +689,7 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -712,7 +711,7 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -734,7 +733,7 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -750,15 +749,13 @@ impl DefaultPlanner {
                             Context::Default => {
                                 let function = NoEq::new(Arc::new(move |s: &mut [Series]| {
                                     let s = std::mem::take(&mut s[0]);
-                                    let len = s.len() as f64;
-                                    parallel_op_series(|s| Ok(s.sum_as_series()), s, None)
-                                        .map(|s| s.cast(&DataType::Float64).unwrap() / len)
+                                    Ok(s.mean_as_series())
                                 })
                                     as Arc<dyn SeriesUdf>);
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -780,7 +777,7 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -802,7 +799,7 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -824,7 +821,7 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -846,7 +843,7 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -864,7 +861,7 @@ impl DefaultPlanner {
                                 let function = NoEq::new(Arc::new(move |s: &mut [Series]| {
                                     let s = std::mem::take(&mut s[0]);
                                     s.n_unique().map(|count| {
-                                        UInt32Chunked::new_from_slice(s.name(), &[count as u32])
+                                        UInt32Chunked::from_slice(s.name(), &[count as u32])
                                             .into_series()
                                     })
                                 })
@@ -872,7 +869,7 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -899,7 +896,7 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -926,14 +923,14 @@ impl DefaultPlanner {
                                 let function = NoEq::new(Arc::new(move |s: &mut [Series]| {
                                     let s = std::mem::take(&mut s[0]);
                                     let count = s.len();
-                                    Ok(UInt32Chunked::new_from_slice(s.name(), &[count as u32])
+                                    Ok(UInt32Chunked::from_slice(s.name(), &[count as u32])
                                         .into_series())
                                 })
                                     as Arc<dyn SeriesUdf>);
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    expr: node_to_exp(expression, expr_arena),
+                                    expr: node_to_expr(expression, expr_arena),
                                     collect_groups: ApplyOptions::ApplyFlat,
                                     auto_explode: false,
                                 }))
@@ -951,7 +948,7 @@ impl DefaultPlanner {
                 Ok(Arc::new(CastExpr {
                     input: phys_expr,
                     data_type,
-                    expr: node_to_exp(expression, expr_arena),
+                    expr: node_to_expr(expression, expr_arena),
                     strict,
                 }))
             }
@@ -967,7 +964,7 @@ impl DefaultPlanner {
                     predicate,
                     truthy,
                     falsy,
-                    expr: node_to_exp(expression, expr_arena),
+                    expr: node_to_expr(expression, expr_arena),
                 }))
             }
             Function {
@@ -981,33 +978,17 @@ impl DefaultPlanner {
                 Ok(Arc::new(ApplyExpr {
                     inputs: input,
                     function,
-                    expr: node_to_exp(expression, expr_arena),
+                    expr: node_to_expr(expression, expr_arena),
                     collect_groups: options.collect_groups,
                     auto_explode: options.auto_explode,
                 }))
             }
-            BinaryFunction {
-                input_a,
-                input_b,
-                function,
-                output_field,
-            } => {
-                let input_a = self.create_physical_expr(input_a, ctxt, expr_arena)?;
-                let input_b = self.create_physical_expr(input_b, ctxt, expr_arena)?;
-                Ok(Arc::new(BinaryFunctionExpr {
-                    input_a,
-                    input_b,
-                    function,
-                    output_field,
-                    expr: node_to_exp(expression, expr_arena),
-                }))
-            }
             Shift { input, periods } => {
                 let input = self.create_physical_expr(input, ctxt, expr_arena)?;
                 Ok(Arc::new(ShiftExpr {
                     input,
                     periods,
-                    expr: node_to_exp(expression, expr_arena),
+                    expr: node_to_expr(expression, expr_arena),
                 }))
             }
             Slice {
@@ -1016,11 +997,13 @@ impl DefaultPlanner {
                 length,
             } => {
                 let input = self.create_physical_expr(input, ctxt, expr_arena)?;
+                let offset = self.create_physical_expr(offset, ctxt, expr_arena)?;
+                let length = self.create_physical_expr(length, ctxt, expr_arena)?;
                 Ok(Arc::new(SliceExpr {
                     input,
                     offset,
-                    len: length,
-                    expr: node_to_exp(expression, expr_arena),
+                    length,
+                    expr: node_to_expr(expression, expr_arena),
                 }))
             }
             Reverse(expr) => {
@@ -1032,7 +1015,7 @@ impl DefaultPlanner {
                 Ok(Arc::new(ApplyExpr {
                     inputs: vec![input],
                     function,
-                    expr: node_to_exp(expression, expr_arena),
+                    expr: node_to_expr(expression, expr_arena),
                     collect_groups: ApplyOptions::ApplyGroups,
                     auto_explode: false,
                 }))
@@ -1046,7 +1029,7 @@ impl DefaultPlanner {
                 Ok(Arc::new(ApplyExpr {
                     inputs: vec![input],
                     function,
-                    expr: node_to_exp(expression, expr_arena),
+                    expr: node_to_expr(expression, expr_arena),
                     collect_groups: ApplyOptions::ApplyGroups,
                     auto_explode: false,
                 }))
@@ -1060,7 +1043,7 @@ impl DefaultPlanner {
                 Ok(Arc::new(ApplyExpr {
                     inputs: vec![input],
                     function,
-                    expr: node_to_exp(expression, expr_arena),
+                    expr: node_to_expr(expression, expr_arena),
                     collect_groups: ApplyOptions::ApplyGroups,
                     auto_explode: false,
                 }))
@@ -1074,12 +1057,13 @@ impl DefaultPlanner {
                 Ok(Arc::new(ApplyExpr {
                     inputs: vec![input],
                     function,
-                    expr: node_to_exp(expression, expr_arena),
+                    expr: node_to_expr(expression, expr_arena),
                     collect_groups: ApplyOptions::ApplyFlat,
                     auto_explode: false,
                 }))
             }
             Wildcard => panic!("should be no wildcard at this point"),
+            Nth(_) => panic!("should be no nth at this point"),
         }
     }
 }
