@@ -2,29 +2,6 @@ use super::*;
 use polars_arrow::prelude::QuantileInterpolOptions;
 use polars_core::series::ops::NullBehavior;
 
-pub(crate) fn load_df() -> DataFrame {
-    df!("a" => &[1, 2, 3, 4, 5],
-                 "b" => &["a", "a", "b", "c", "c"],
-                 "c" => &[1, 2, 3, 4, 5]
-    )
-    .unwrap()
-}
-
-#[test]
-fn test_lazy_ternary() {
-    let df = get_df()
-        .lazy()
-        .with_column(
-            when(col("sepal.length").lt(lit(5.0)))
-                .then(lit(10))
-                .otherwise(lit(1))
-                .alias("new"),
-        )
-        .collect()
-        .unwrap();
-    assert_eq!(Some(43), df.column("new").unwrap().sum::<i32>());
-}
-
 #[test]
 fn test_lazy_with_column() {
     let df = get_df()
@@ -52,7 +29,7 @@ fn test_lazy_exec() {
         .clone()
         .lazy()
         .select([col("sepal.width"), col("variety")])
-        .sort("sepal.width", false)
+        .sort("sepal.width", Default::default())
         .collect();
     println!("{:?}", new);
 
@@ -91,7 +68,6 @@ fn test_lazy_melt() {
         .collect()
         .unwrap();
     assert_eq!(out.shape(), (7, 3));
-    dbg!(out);
 }
 
 #[test]
@@ -199,18 +175,16 @@ fn test_lazy_agg() {
         .lazy()
         .groupby([col("date")])
         .agg([
-            col("rain").min(),
-            col("rain").sum(),
+            col("rain").min().alias("min"),
+            col("rain").sum().alias("sum"),
             col("rain")
                 .quantile(0.5, QuantileInterpolOptions::default())
                 .alias("median_rain"),
         ])
-        .sort("date", false);
+        .sort("date", Default::default());
 
-    println!("{:?}", lf.describe_plan());
-    println!("{:?}", lf.describe_optimized_plan());
     let new = lf.collect().unwrap();
-    println!("{:?}", new);
+    dbg!(new);
 }
 
 #[test]
@@ -288,7 +262,7 @@ fn test_lazy_query_1() {
         .filter(col("a").lt(lit(2)))
         .groupby([col("b")])
         .agg([col("b").first(), col("c").first()])
-        .select([col("b"), col("c_first")])
+        .select([col("b"), col("c")])
         .collect()
         .unwrap();
 }
@@ -390,29 +364,6 @@ fn test_lazy_query_5() {
     assert_eq!(s.len(), 2);
 }
 
-#[test]
-fn test_lazy_query_6() -> Result<()> {
-    let df = df! {
-        "uid" => [0, 0, 0, 1, 1, 1],
-        "day" => [1, 2, 4, 1, 2, 3],
-        "cumcases" => [10, 12, 15, 25, 30, 41]
-    }
-    .unwrap();
-
-    let out = df
-        .lazy()
-        .groupby([col("uid")])
-        // a double aggregation expression.
-        .agg([pearson_corr(col("day"), col("cumcases")).pow(2.0)])
-        .sort("uid", false)
-        .collect()
-        .unwrap();
-    let s = out.column("pearson_corr")?.f64()?;
-    assert!((s.get(0).unwrap() - 0.994360902255639).abs() < 0.000001);
-    assert!((s.get(1).unwrap() - 0.9552238805970149).abs() < 0.000001);
-    Ok(())
-}
-
 #[test]
 #[cfg(feature = "is_in")]
 fn test_lazy_query_8() -> Result<()> {
@@ -473,7 +424,7 @@ fn test_lazy_query_9() -> Result<()> {
         )
         .groupby([col("Cities.Country")])
         .agg([col("Sales.Amount").sum().alias("sum")])
-        .sort("sum", false)
+        .sort("sum", Default::default())
         .collect()?;
     let vals = out
         .column("sum")?
@@ -485,13 +436,18 @@ fn test_lazy_query_9() -> Result<()> {
 }
 
 #[test]
-#[cfg(feature = "temporal")]
+#[cfg(all(
+    feature = "temporal",
+    feature = "dtype-datetime",
+    feature = "dtype-date",
+    feature = "dtype-duration"
+))]
 fn test_lazy_query_10() {
     use polars_core::export::chrono::Duration as ChronoDuration;
     let date = NaiveDate::from_ymd(2021, 3, 5);
-    let x: Series = DatetimeChunked::new_from_naive_datetime(
+    let x: Series = DatetimeChunked::from_naive_datetime(
         "x",
-        &*vec![
+        [
             NaiveDateTime::new(date, NaiveTime::from_hms(12, 0, 0)),
             NaiveDateTime::new(date, NaiveTime::from_hms(13, 0, 0)),
             NaiveDateTime::new(date, NaiveTime::from_hms(14, 0, 0)),
@@ -499,9 +455,9 @@ fn test_lazy_query_10() {
         TimeUnit::Nanoseconds,
     )
     .into();
-    let y: Series = DatetimeChunked::new_from_naive_datetime(
+    let y: Series = DatetimeChunked::from_naive_datetime(
         "y",
-        &*vec![
+        [
             NaiveDateTime::new(date, NaiveTime::from_hms(11, 0, 0)),
             NaiveDateTime::new(date, NaiveTime::from_hms(11, 0, 0)),
             NaiveDateTime::new(date, NaiveTime::from_hms(11, 0, 0)),
@@ -515,9 +471,9 @@ fn test_lazy_query_10() {
         .select(&[(col("x") - col("y")).alias("z")])
         .collect()
         .unwrap();
-    let z: Series = DurationChunked::new_from_duration(
+    let z: Series = DurationChunked::from_duration(
         "z",
-        &*vec![
+        [
             ChronoDuration::hours(1),
             ChronoDuration::hours(2),
             ChronoDuration::hours(3),
@@ -526,9 +482,9 @@ fn test_lazy_query_10() {
     )
     .into();
     assert!(out.column("z").unwrap().series_equal(&z));
-    let x: Series = DatetimeChunked::new_from_naive_datetime(
+    let x: Series = DatetimeChunked::from_naive_datetime(
         "x",
-        &*vec![
+        [
             NaiveDateTime::new(date, NaiveTime::from_hms(2, 0, 0)),
             NaiveDateTime::new(date, NaiveTime::from_hms(3, 0, 0)),
             NaiveDateTime::new(date, NaiveTime::from_hms(4, 0, 0)),
@@ -536,9 +492,9 @@ fn test_lazy_query_10() {
         TimeUnit::Milliseconds,
     )
     .into();
-    let y: Series = DatetimeChunked::new_from_naive_datetime(
+    let y: Series = DatetimeChunked::from_naive_datetime(
         "y",
-        &*vec![
+        [
             NaiveDateTime::new(date, NaiveTime::from_hms(1, 0, 0)),
             NaiveDateTime::new(date, NaiveTime::from_hms(1, 0, 0)),
             NaiveDateTime::new(date, NaiveTime::from_hms(1, 0, 0)),
@@ -559,10 +515,14 @@ fn test_lazy_query_10() {
 }
 
 #[test]
-#[cfg(feature = "temporal")]
+#[cfg(all(
+    feature = "temporal",
+    feature = "dtype-date",
+    feature = "dtype-datetime"
+))]
 fn test_lazy_query_7() {
     let date = NaiveDate::from_ymd(2021, 3, 5);
-    let dates = vec![
+    let dates = [
         NaiveDateTime::new(date, NaiveTime::from_hms(12, 0, 0)),
         NaiveDateTime::new(date, NaiveTime::from_hms(12, 1, 0)),
         NaiveDateTime::new(date, NaiveTime::from_hms(12, 2, 0)),
@@ -572,7 +532,7 @@ fn test_lazy_query_7() {
     ];
     let data = vec![Some(1.), Some(2.), Some(3.), Some(4.), None, None];
     let df = DataFrame::new(vec![
-        DatetimeChunked::new_from_naive_datetime("date", &*dates, TimeUnit::Nanoseconds).into(),
+        DatetimeChunked::from_naive_datetime("date", dates, TimeUnit::Nanoseconds).into(),
         Series::new("data", data),
     ])
     .unwrap();
@@ -632,7 +592,7 @@ fn test_simplify_expr() {
     let rules: &mut [Box<dyn OptimizationRule>] = &mut [Box::new(SimplifyExprRule {})];
 
     let optimizer = StackOptimizer {};
-    let mut lp_top = to_alp(plan, &mut expr_arena, &mut lp_arena);
+    let mut lp_top = to_alp(plan, &mut expr_arena, &mut lp_arena).unwrap();
     lp_top = optimizer.optimize_loop(rules, &mut expr_arena, &mut lp_arena, lp_top);
     let plan = node_to_lp(lp_top, &mut expr_arena, &mut lp_arena);
     assert!(
@@ -667,48 +627,6 @@ fn test_lazy_reverse() {
         .frame_equal_missing(&df.reverse()))
 }
 
-#[test]
-fn test_lazy_agg_scan() {
-    let lf = scan_foods_csv;
-    let df = lf().min().collect().unwrap();
-    assert!(df.frame_equal_missing(&lf().collect().unwrap().min()));
-    let df = lf().max().collect().unwrap();
-    assert!(df.frame_equal_missing(&lf().collect().unwrap().max()));
-    // mean is not yet aggregated at scan.
-    let df = lf().mean().collect().unwrap();
-    assert!(df.frame_equal_missing(&lf().collect().unwrap().mean()));
-}
-
-#[test]
-fn test_lazy_df_aggregations() {
-    let df = load_df();
-
-    assert!(df
-        .clone()
-        .lazy()
-        .min()
-        .collect()
-        .unwrap()
-        .frame_equal_missing(&df.min()));
-    assert!(df
-        .clone()
-        .lazy()
-        .median()
-        .collect()
-        .unwrap()
-        .frame_equal_missing(&df.median()));
-    assert!(df
-        .clone()
-        .lazy()
-        .quantile(0.5, QuantileInterpolOptions::default())
-        .collect()
-        .unwrap()
-        .frame_equal_missing(
-            &df.quantile(0.5, QuantileInterpolOptions::default())
-                .unwrap()
-        ));
-}
-
 #[test]
 fn test_lazy_predicate_pushdown_binary_expr() {
     let df = load_df();
@@ -742,43 +660,6 @@ fn test_lazy_fill_null() {
     assert_eq!(out.get_column_names(), vec!["a", "b"])
 }
 
-#[test]
-fn test_lazy_window_functions() {
-    let df = df! {
-        "groups" => &[1, 1, 2, 2, 1, 2, 3, 3, 1],
-        "values" => &[1, 2, 3, 4, 5, 6, 7, 8, 8]
-    }
-    .unwrap();
-
-    // sums
-    // 1 => 16
-    // 2 => 13
-    // 3 => 15
-    let correct = [16, 16, 13, 13, 16, 13, 15, 15, 16]
-        .iter()
-        .copied()
-        .map(Some)
-        .collect::<Vec<_>>();
-
-    // test if groups is available after projection pushdown.
-    let _ = df
-        .clone()
-        .lazy()
-        .select(&[avg("values").over([col("groups")]).alias("part")])
-        .collect()
-        .unwrap();
-    // test if partition aggregation is correct
-    let out = df
-        .lazy()
-        .select([col("groups"), sum("values").over([col("groups")])])
-        .collect()
-        .unwrap();
-    assert_eq!(
-        Vec::from(out.select_at_idx(1).unwrap().i32().unwrap()),
-        correct
-    );
-}
-
 #[test]
 fn test_lazy_double_projection() {
     let df = df! {
@@ -807,7 +688,7 @@ fn test_type_coercion() {
     let rules: &mut [Box<dyn OptimizationRule>] = &mut [Box::new(TypeCoercionRule {})];
 
     let optimizer = StackOptimizer {};
-    let mut lp_top = to_alp(lp, &mut expr_arena, &mut lp_arena);
+    let mut lp_top = to_alp(lp, &mut expr_arena, &mut lp_arena).unwrap();
     lp_top = optimizer.optimize_loop(rules, &mut expr_arena, &mut lp_arena, lp_top);
     let lp = node_to_lp(lp_top, &mut expr_arena, &mut lp_arena);
 
@@ -834,26 +715,24 @@ fn test_lazy_partition_agg() {
         .lazy()
         .groupby([col("foo")])
         .agg([col("bar").mean()])
-        .sort("foo", false)
+        .sort("foo", Default::default())
         .collect()
         .unwrap();
 
     assert_eq!(
-        Vec::from(out.column("bar_mean").unwrap().f64().unwrap()),
+        Vec::from(out.column("bar").unwrap().f64().unwrap()),
         &[Some(1.0), Some(2.0), Some(3.0)]
     );
 
     let out = scan_foods_csv()
         .groupby([col("category")])
         .agg([col("calories").list()])
-        .sort("category", false)
+        .sort("category", Default::default())
         .collect()
         .unwrap();
-    dbg!(&out);
     let cat_agg_list = out.select_at_idx(1).unwrap();
     let fruit_series = cat_agg_list.list().unwrap().get(0).unwrap();
     let fruit_list = fruit_series.i64().unwrap();
-    dbg!(fruit_list);
     assert_eq!(
         Vec::from(fruit_list),
         &[
@@ -926,14 +805,11 @@ fn test_lazy_groupby() {
         .lazy()
         .groupby([col("groups")])
         .agg([col("a").mean()])
-        .sort("a_mean", false)
+        .sort("a", Default::default())
         .collect()
         .unwrap();
 
-    assert_eq!(
-        out.column("a_mean").unwrap().f64().unwrap().get(0),
-        Some(1.0)
-    );
+    assert_eq!(out.column("a").unwrap().f64().unwrap().get(0), Some(1.0));
 }
 
 #[test]
@@ -966,7 +842,7 @@ fn test_lazy_groupby_sort() {
         .unwrap();
 
     assert_eq!(
-        Vec::from(out.column("b_first").unwrap().i32().unwrap()),
+        Vec::from(out.column("b").unwrap().i32().unwrap()),
         [Some(1), Some(2), Some(6)]
     );
 
@@ -980,7 +856,7 @@ fn test_lazy_groupby_sort() {
         .unwrap();
 
     assert_eq!(
-        Vec::from(out.column("b_last").unwrap().i32().unwrap()),
+        Vec::from(out.column("b").unwrap().i32().unwrap()),
         [Some(3), Some(5), Some(6)]
     );
 }
@@ -1004,7 +880,7 @@ fn test_lazy_groupby_sort_by() {
         .unwrap();
 
     assert_eq!(
-        Vec::from(out.column("b_first").unwrap().i32().unwrap()),
+        Vec::from(out.column("b").unwrap().i32().unwrap()),
         [Some(1), Some(4), Some(6)]
     );
 }
@@ -1042,11 +918,11 @@ fn test_lazy_groupby_binary_expr() {
         .lazy()
         .groupby([col("a")])
         .agg([col("b").mean() * lit(2)])
-        .sort("a", false)
+        .sort("a", Default::default())
         .collect()
         .unwrap();
     assert_eq!(
-        Vec::from(out.column("b_mean").unwrap().f64().unwrap()),
+        Vec::from(out.column("b").unwrap().f64().unwrap()),
         [Some(4.0), Some(9.0), Some(12.0)]
     );
 }
@@ -1065,12 +941,27 @@ fn test_lazy_groupby_filter() -> Result<()> {
         .lazy()
         .groupby([col("a")])
         .agg([
-            col("b").filter(col("a").eq(lit("a"))).sum(),
-            col("b").filter(col("a").eq(lit("a"))).first(),
-            col("b").filter(col("a").eq(lit("e"))).mean(),
-            col("b").filter(col("a").eq(lit("a"))).last(),
+            col("b").filter(col("a").eq(lit("a"))).sum().alias("b_sum"),
+            col("b")
+                .filter(col("a").eq(lit("a")))
+                .first()
+                .alias("b_first"),
+            col("b")
+                .filter(col("a").eq(lit("e")))
+                .mean()
+                .alias("b_mean"),
+            col("b")
+                .filter(col("a").eq(lit("a")))
+                .last()
+                .alias("b_last"),
         ])
-        .sort("a", false)
+        .sort(
+            "a",
+            SortOptions {
+                descending: false,
+                nulls_last: false,
+            },
+        )
         .collect()?;
 
     dbg!(&out);
@@ -1137,17 +1028,23 @@ fn test_groupby_sort_slice() -> Result<()> {
     let out1 = df
         .clone()
         .lazy()
-        .sort("vals", true)
+        .sort(
+            "vals",
+            SortOptions {
+                descending: true,
+                ..Default::default()
+            },
+        )
         .groupby([col("groups")])
         .agg([col("vals").head(Some(2)).alias("foo")])
-        .sort("groups", false)
+        .sort("groups", SortOptions::default())
         .collect()?;
 
     let out2 = df
         .lazy()
         .groupby([col("groups")])
         .agg([col("vals").sort(true).head(Some(2)).alias("foo")])
-        .sort("groups", false)
+        .sort("groups", SortOptions::default())
         .collect()?;
 
     assert!(out1.column("foo")?.series_equal(out2.column("foo")?));
@@ -1166,7 +1063,7 @@ fn test_groupby_cumsum() -> Result<()> {
         .lazy()
         .groupby([col("groups")])
         .agg([col("vals").cumsum(false)])
-        .sort("groups", false)
+        .sort("groups", Default::default())
         .collect()?;
 
     dbg!(&out);
@@ -1198,7 +1095,7 @@ fn test_argsort_multiple() -> Result<()> {
         .collect()?;
 
     assert_eq!(
-        Vec::from(out.column("int")?.u32()?),
+        Vec::from(out.column("int")?.idx()?),
         [2, 4, 1, 3, 0]
             .iter()
             .copied()
@@ -1250,9 +1147,10 @@ fn test_filter_and_alias() -> Result<()> {
 
     let expected = df![
         "a" => [2, 2],
-        "a_squared" => [4, 4]
+        "a_squared" => [4.0, 4.0]
     ]?;
-
+    println!("{:?}", out);
+    println!("{:?}", expected);
     assert!(out.frame_equal(&expected));
     Ok(())
 }
@@ -1299,7 +1197,7 @@ fn test_fill_forward() -> Result<()> {
 
     let out = df
         .lazy()
-        .select([col("b").forward_fill().over([col("a")])])
+        .select([col("b").forward_fill().list().over([col("a")])])
         .collect()?;
     let agg = out.column("b")?.list()?;
 
@@ -1438,7 +1336,9 @@ fn test_filter_in_groupby_agg() -> Result<()> {
         .clone()
         .lazy()
         .groupby([col("a")])
-        .agg([(col("b").filter(col("b").eq(lit(100))) * lit(2)).mean()])
+        .agg([(col("b").filter(col("b").eq(lit(100))) * lit(2))
+            .mean()
+            .alias("b_mean")])
         .collect()?;
 
     assert_eq!(out.column("b_mean")?.null_count(), 2);
@@ -1449,223 +1349,14 @@ fn test_filter_in_groupby_agg() -> Result<()> {
         .agg([(col("b")
             .filter(col("b").eq(lit(100)))
             .map(Ok, GetOutput::same_type()))
-        .mean()])
+        .mean()
+        .alias("b_mean")])
         .collect()?;
     assert_eq!(out.column("b_mean")?.null_count(), 2);
 
     Ok(())
 }
 
-#[test]
-fn test_shift_and_fill_window_function() -> Result<()> {
-    let df = fruits_cars();
-
-    // a ternary expression with a final list aggregation
-    let out1 = df
-        .clone()
-        .lazy()
-        .select([
-            col("fruits"),
-            col("B").shift_and_fill(-1, lit(-1)).over([col("fruits")]),
-        ])
-        .collect()?;
-
-    // same expression, no final list aggregation
-    let out2 = df
-        .lazy()
-        .select([
-            col("fruits"),
-            col("B")
-                .shift_and_fill(-1, lit(-1))
-                .list()
-                .over([col("fruits")]),
-        ])
-        .collect()?;
-
-    dbg!(&out1, &out2);
-
-    assert!(out1.frame_equal(&out2));
-
-    Ok(())
-}
-
-#[test]
-fn test_cumsum_agg_as_key() -> Result<()> {
-    let df = df![
-        "depth" => &[0i32, 1, 2, 3, 4, 5, 6, 7, 8, 9],
-        "soil" => &["peat", "peat", "peat", "silt", "silt", "silt", "sand", "sand", "peat", "peat"]
-    ]?;
-    // this checks if the grouper can work with the complex query as a key
-
-    let out = df
-        .lazy()
-        .groupby([col("soil")
-            .neq(col("soil").shift_and_fill(1, col("soil").first()))
-            .cumsum(false)
-            .alias("key")])
-        .agg([col("depth").max().keep_name()])
-        .sort("depth", false)
-        .collect()?;
-
-    assert_eq!(
-        Vec::from(out.column("key")?.u32()?),
-        &[Some(0), Some(1), Some(2), Some(3)]
-    );
-    assert_eq!(
-        Vec::from(out.column("depth")?.i32()?),
-        &[Some(2), Some(5), Some(7), Some(9)]
-    );
-
-    Ok(())
-}
-
-#[test]
-fn test_auto_list_agg() -> Result<()> {
-    let df = fruits_cars();
-
-    // test if alias executor adds a list after shift and fill
-    let out = df
-        .clone()
-        .lazy()
-        .groupby([col("fruits")])
-        .agg([col("B").shift_and_fill(-1, lit(-1)).alias("foo")])
-        .collect()?;
-
-    assert!(matches!(out.column("foo")?.dtype(), DataType::List(_)));
-
-    // test if it runs and groupby executor thus implements a list after shift_and_fill
-    let _out = df
-        .clone()
-        .lazy()
-        .groupby([col("fruits")])
-        .agg([col("B").shift_and_fill(-1, lit(-1))])
-        .collect()?;
-
-    // test if window expr executor adds list
-    let _out = df
-        .clone()
-        .lazy()
-        .select([col("B").shift_and_fill(-1, lit(-1)).alias("foo")])
-        .collect()?;
-
-    let _out = df
-        .lazy()
-        .select([col("B").shift_and_fill(-1, lit(-1))])
-        .collect()?;
-    Ok(())
-}
-
-#[test]
-fn test_exploded_window_function() -> Result<()> {
-    let df = fruits_cars();
-
-    let out = df
-        .clone()
-        .lazy()
-        .sort("fruits", false)
-        .select([
-            col("fruits"),
-            col("B")
-                .shift(1)
-                .over([col("fruits")])
-                .explode()
-                .alias("shifted"),
-        ])
-        .collect()?;
-
-    assert_eq!(
-        Vec::from(out.column("shifted")?.i32()?),
-        &[None, Some(3), None, Some(5), Some(4)]
-    );
-
-    // this tests if cast succeeds in aggregation context
-    // we implicitly also test that a literal does not upcast a column
-    let out = df
-        .lazy()
-        .sort("fruits", false)
-        .select([
-            col("fruits"),
-            col("B")
-                .shift_and_fill(1, lit(-1.0f32))
-                .over([col("fruits")])
-                .explode()
-                .alias("shifted"),
-        ])
-        .collect()?;
-
-    assert_eq!(
-        Vec::from(out.column("shifted")?.f32()?),
-        &[Some(-1.0), Some(3.0), Some(-1.0), Some(5.0), Some(4.0)]
-    );
-    Ok(())
-}
-
-#[test]
-fn test_reverse_in_groups() -> Result<()> {
-    let df = fruits_cars();
-
-    let out = df
-        .lazy()
-        .sort("fruits", false)
-        .select([col("B")
-            .reverse()
-            .over([col("fruits")])
-            .explode()
-            .alias("rev")])
-        .collect()?;
-
-    assert_eq!(
-        Vec::from(out.column("rev")?.i32()?),
-        &[Some(2), Some(3), Some(1), Some(4), Some(5)]
-    );
-    Ok(())
-}
-
-#[test]
-fn test_take_in_groups() -> Result<()> {
-    let df = fruits_cars();
-
-    let out = df
-        .lazy()
-        .sort("fruits", false)
-        .select([col("B")
-            .take(lit(Series::new("", &[0u32])))
-            .over([col("fruits")])
-            .alias("taken")])
-        .collect()?;
-
-    assert_eq!(
-        Vec::from(out.column("taken")?.i32()?),
-        &[Some(3), Some(3), Some(5), Some(5), Some(5)]
-    );
-    Ok(())
-}
-
-#[test]
-fn test_sort_by_in_groups() -> Result<()> {
-    let df = fruits_cars();
-
-    let out = df
-        .lazy()
-        .sort("cars", false)
-        .select([
-            col("fruits"),
-            col("cars"),
-            col("A")
-                .sort_by([col("B")], [false])
-                .over([col("cars")])
-                .explode()
-                .alias("sorted_A_by_B"),
-        ])
-        .collect()?;
-
-    assert_eq!(
-        Vec::from(out.column("sorted_A_by_B")?.i32()?),
-        &[Some(2), Some(5), Some(4), Some(3), Some(1)]
-    );
-    Ok(())
-}
-
 #[test]
 fn test_sort_by() -> Result<()> {
     let df = df![
@@ -1707,7 +1398,7 @@ fn test_sort_by() -> Result<()> {
         .agg([col("a").sort_by([col("b"), col("c")], [false]).list()])
         .collect()?;
 
-    let a = out.column("a_agg_list")?.explode()?;
+    let a = out.column("a")?.explode()?;
     assert_eq!(
         Vec::from(a.i32().unwrap()),
         &[Some(3), Some(1), Some(2), Some(5), Some(4)]
@@ -1727,11 +1418,11 @@ fn test_filter_after_shift_in_groups() -> Result<()> {
             col("B")
                 .shift(1)
                 .filter(col("B").shift(1).gt(lit(4)))
+                .list()
                 .over([col("fruits")])
                 .alias("filtered"),
         ])
         .collect()?;
-    dbg!(out.column("filtered")?);
 
     assert_eq!(
         out.column("filtered")?
@@ -1788,8 +1479,8 @@ fn test_categorical_addition() -> Result<()> {
     let out = df
         .lazy()
         .select([
-            col("fruits").cast(DataType::Categorical),
-            col("cars").cast(DataType::Categorical),
+            col("fruits").cast(DataType::Categorical(None)),
+            col("cars").cast(DataType::Categorical(None)),
         ])
         .select([(col("fruits") + lit(" ") + col("cars")).alias("foo")])
         .collect()?;
@@ -1831,7 +1522,13 @@ fn test_groupby_small_ints() -> Result<()> {
         .lazy()
         .groupby([col("id_16"), col("id_32")])
         .agg([col("id_16").sum().alias("foo")])
-        .sort("foo", true)
+        .sort(
+            "foo",
+            SortOptions {
+                descending: true,
+                ..Default::default()
+            },
+        )
         .collect()?;
 
     assert_eq!(Vec::from(out.column("foo")?.i64()?), &[Some(2), Some(1)]);
@@ -1848,7 +1545,7 @@ fn test_when_then_schema() -> Result<()> {
             .then(Null {}.lit())
             .otherwise(col("A"))])
         .schema();
-    assert_ne!(schema.fields()[0].data_type(), &DataType::Null);
+    assert_ne!(schema.get_index(0).unwrap().1, &DataType::Null);
 
     Ok(())
 }
@@ -1872,6 +1569,7 @@ fn test_sort_by_suffix() -> Result<()> {
         .lazy()
         .select([col("*")
             .sort_by([col("A")], [false])
+            .list()
             .over([col("fruits")])
             .flatten()
             .suffix("_sorted")])
@@ -1964,60 +1662,6 @@ fn test_round_after_agg() -> Result<()> {
     Ok(())
 }
 
-#[test]
-fn test_power_in_agg_list1() -> Result<()> {
-    let df = fruits_cars();
-
-    // this test if the group tuples are correctly updated after
-    // a flat apply on a final aggregation
-    let out = df
-        .lazy()
-        .groupby([col("fruits")])
-        .agg([col("A")
-            .rolling_min(RollingOptions {
-                window_size: 1,
-                ..Default::default()
-            })
-            .pow(2.0)
-            .alias("foo")])
-        .sort("fruits", true)
-        .collect()?;
-
-    let agg = out.column("foo")?.list()?;
-    let first = agg.get(0).unwrap();
-    let vals = first.f64()?;
-    assert_eq!(Vec::from(vals), &[Some(1.0), Some(4.0), Some(25.0)]);
-
-    Ok(())
-}
-
-#[test]
-fn test_power_in_agg_list2() -> Result<()> {
-    let df = fruits_cars();
-
-    // this test if the group tuples are correctly updated after
-    // a flat apply on evaluate_on_groups
-    let out = df
-        .lazy()
-        .groupby([col("fruits")])
-        .agg([col("A")
-            .rolling_min(RollingOptions {
-                window_size: 2,
-                min_periods: 2,
-                ..Default::default()
-            })
-            .pow(2.0)
-            .sum()
-            .alias("foo")])
-        .sort("fruits", true)
-        .collect()?;
-
-    let agg = out.column("foo")?.f64()?;
-    assert_eq!(Vec::from(agg), &[Some(5.0), Some(9.0)]);
-
-    Ok(())
-}
-
 #[test]
 #[cfg(feature = "dtype-date")]
 fn test_fill_nan() -> Result<()> {
@@ -2032,26 +1676,6 @@ fn test_fill_nan() -> Result<()> {
     Ok(())
 }
 
-#[test]
-fn test_agg_exprs() -> Result<()> {
-    let df = fruits_cars();
-
-    // a binary expression followed by a function and an aggregation. See if it runs
-    let out = df
-        .lazy()
-        .groupby_stable([col("cars")])
-        .agg([(lit(1) - col("A"))
-            .map(|s| Ok(&s * 2), GetOutput::same_type())
-            .list()
-            .alias("foo")])
-        .collect()?;
-    let ca = out.column("foo")?.list()?;
-    let out = ca.lst_lengths();
-
-    assert_eq!(Vec::from(&out), &[Some(4), Some(1)]);
-    Ok(())
-}
-
 #[test]
 fn test_exclude_regex() -> Result<()> {
     let df = fruits_cars();
@@ -2078,7 +1702,7 @@ fn test_groupby_rank() -> Result<()> {
 
     let out = out.column("B")?;
     let out = out.list()?.get(1).unwrap();
-    let out = out.u32()?;
+    let out = out.idx()?;
 
     assert_eq!(Vec::from(out), &[Some(1)]);
     Ok(())
@@ -2093,7 +1717,7 @@ fn test_apply_multiple_columns() -> Result<()> {
     let out = df
         .clone()
         .lazy()
-        .select([map_mul(
+        .select([map_multiple(
             multiply,
             [col("A"), col("B")],
             GetOutput::from_type(DataType::Float64),
@@ -2109,7 +1733,7 @@ fn test_apply_multiple_columns() -> Result<()> {
     let out = df
         .lazy()
         .groupby_stable([col("cars")])
-        .agg([apply_mul(
+        .agg([apply_multiple(
             multiply,
             [col("A"), col("B")],
             GetOutput::from_type(DataType::Float64),
@@ -2192,52 +1816,6 @@ fn test_drop_and_select() -> Result<()> {
     Ok(())
 }
 
-#[test]
-fn test_take_consistency() -> Result<()> {
-    let df = fruits_cars();
-    let out = df
-        .clone()
-        .lazy()
-        .select([col("A").arg_sort(true).take(lit(0))])
-        .collect()?;
-
-    assert_eq!(out.column("A")?.get(0), AnyValue::UInt32(4));
-
-    let out = df
-        .clone()
-        .lazy()
-        .groupby_stable([col("cars")])
-        .agg([col("A").arg_sort(true).take(lit(0))])
-        .collect()?;
-
-    let out = out.column("A")?;
-    let out = out.u32()?;
-    assert_eq!(Vec::from(out), &[Some(3), Some(0)]);
-
-    let out_df = df
-        .clone()
-        .lazy()
-        .groupby_stable([col("cars")])
-        .agg([
-            col("A"),
-            col("A").arg_sort(true).take(lit(0)).alias("1"),
-            col("A")
-                .take(col("A").arg_sort(true).take(lit(0)))
-                .alias("2"),
-        ])
-        .collect()?;
-
-    let out = out_df.column("2")?;
-    let out = out.i32()?;
-    assert_eq!(Vec::from(out), &[Some(5), Some(2)]);
-
-    let out = out_df.column("1")?;
-    let out = out.u32()?;
-    assert_eq!(Vec::from(out), &[Some(3), Some(0)]);
-
-    Ok(())
-}
-
 #[test]
 fn test_groupby_on_lists() -> Result<()> {
     let s0 = Series::new("", [1i32, 2, 3]);
@@ -2257,7 +1835,7 @@ fn test_groupby_on_lists() -> Result<()> {
         .collect()?;
 
     assert_eq!(
-        out.column("arrays_first")?.dtype(),
+        out.column("arrays")?.dtype(),
         &DataType::List(Box::new(DataType::Int32))
     );
 
@@ -2269,7 +1847,7 @@ fn test_groupby_on_lists() -> Result<()> {
         .collect()?;
 
     assert_eq!(
-        out.column("arrays_agg_list")?.dtype(),
+        out.column("arrays")?.dtype(),
         &DataType::List(Box::new(DataType::List(Box::new(DataType::Int32))))
     );
 
@@ -2286,180 +1864,15 @@ fn test_single_group_result() -> Result<()> {
 
     let out = df
         .lazy()
-        .select([col("a").arg_sort(false).over([col("a")]).flatten()])
+        .select([col("a").arg_sort(false).list().over([col("a")]).flatten()])
         .collect()?;
 
-    let a = out.column("a")?.u32()?;
+    let a = out.column("a")?.idx()?;
     assert_eq!(Vec::from(a), &[Some(0), Some(0)]);
 
     Ok(())
 }
 
-#[test]
-fn test_literal_window_fn() -> Result<()> {
-    let df = df![
-        "chars" => ["a", "a", "b"]
-    ]?;
-
-    let out = df
-        .lazy()
-        .select([lit(1).cumsum(false).over([col("chars")]).alias("foo")])
-        .collect()?;
-
-    let out = out.column("foo")?;
-    assert!(matches!(out.dtype(), DataType::List(_)));
-    let flat = out.explode()?;
-    let flat = flat.i32()?;
-    assert_eq!(
-        Vec::from(flat),
-        &[Some(1), Some(2), Some(1), Some(2), Some(1)]
-    );
-
-    Ok(())
-}
-
-#[test]
-fn test_binary_agg_context_0() -> Result<()> {
-    let df = df![
-        "groups" => [1, 1, 2, 2, 3, 3],
-        "vals" => [1, 2, 3, 4, 5, 6]
-    ]
-    .unwrap();
-
-    let out = df
-        .lazy()
-        .groupby_stable([col("groups")])
-        .agg([when(col("vals").first().neq(lit(1)))
-            .then(lit("a"))
-            .otherwise(lit("b"))
-            .alias("foo")])
-        .collect()
-        .unwrap();
-
-    let out = out.column("foo")?;
-    let out = out.explode()?;
-    let out = out.utf8()?;
-    assert_eq!(
-        Vec::from(out),
-        &[
-            Some("b"),
-            Some("b"),
-            Some("a"),
-            Some("a"),
-            Some("a"),
-            Some("a")
-        ]
-    );
-    Ok(())
-}
-
-// just like binary expression, this must be changed. This can work
-#[test]
-fn test_binary_agg_context_1() -> Result<()> {
-    let df = df![
-        "groups" => [1, 1, 2, 2, 3, 3],
-        "vals" => [1, 13, 3, 87, 1, 6]
-    ]?;
-
-    // groups
-    // 1 => [1, 13]
-    // 2 => [3, 87]
-    // 3 => [1, 6]
-
-    let out = df
-        .clone()
-        .lazy()
-        .groupby_stable([col("groups")])
-        .agg([when(col("vals").eq(lit(1)))
-            .then(col("vals").sum())
-            .otherwise(lit(90))
-            .alias("vals")])
-        .collect()?;
-
-    // if vals == 1 then sum(vals) else vals
-    // [14, 90]
-    // [90, 90]
-    // [7, 90]
-    let out = out.column("vals")?;
-    let out = out.explode()?;
-    let out = out.i32()?;
-    assert_eq!(
-        Vec::from(out),
-        &[Some(14), Some(90), Some(90), Some(90), Some(7), Some(90)]
-    );
-
-    let out = df
-        .lazy()
-        .groupby_stable([col("groups")])
-        .agg([when(col("vals").eq(lit(1)))
-            .then(lit(90))
-            .otherwise(col("vals").sum())
-            .alias("vals")])
-        .collect()?;
-
-    // if vals == 1 then 90 else sum(vals)
-    // [90, 14]
-    // [90, 90]
-    // [90, 7]
-    let out = out.column("vals")?;
-    let out = out.explode()?;
-    let out = out.i32()?;
-    assert_eq!(
-        Vec::from(out),
-        &[Some(90), Some(14), Some(90), Some(90), Some(90), Some(7)]
-    );
-
-    Ok(())
-}
-
-#[test]
-fn test_binary_agg_context_2() -> Result<()> {
-    let df = df![
-        "groups" => [1, 1, 2, 2, 3, 3],
-        "vals" => [1, 2, 3, 4, 5, 6]
-    ]?;
-
-    // this is complex because we first aggregate one expression of the binary operation.
-
-    let out = df
-        .clone()
-        .lazy()
-        .groupby_stable([col("groups")])
-        .agg([((col("vals").first() - col("vals")).list()).alias("vals")])
-        .collect()?;
-
-    // 0 - [1, 2] = [0, -1]
-    // 3 - [3, 4] = [0, -1]
-    // 5 - [5, 6] = [0, -1]
-    let out = out.column("vals")?;
-    let out = out.explode()?;
-    let out = out.i32()?;
-    assert_eq!(
-        Vec::from(out),
-        &[Some(0), Some(-1), Some(0), Some(-1), Some(0), Some(-1)]
-    );
-
-    // Same, but now we reverse the lhs / rhs.
-    let out = df
-        .lazy()
-        .groupby_stable([col("groups")])
-        .agg([((col("vals")) - col("vals").first()).list().alias("vals")])
-        .collect()?;
-
-    // [1, 2] - 1 = [0, 1]
-    // [3, 4] - 3 = [0, 1]
-    // [5, 6] - 5 = [0, 1]
-    let out = out.column("vals")?;
-    let out = out.explode()?;
-    let out = out.i32()?;
-    assert_eq!(
-        Vec::from(out),
-        &[Some(0), Some(1), Some(0), Some(1), Some(0), Some(1)]
-    );
-
-    Ok(())
-}
-
 #[test]
 fn test_single_ranked_group() -> Result<()> {
     // tests type consistency of rank algorithm
@@ -2474,6 +1887,7 @@ fn test_single_ranked_group() -> Result<()> {
                 method: RankMethod::Average,
                 ..Default::default()
             })
+            .list()
             .over([col("group")])])
         .collect()?;
 
@@ -2518,7 +1932,7 @@ fn test_apply_flatten() -> Result<()> {
     let out = df
         .lazy()
         .groupby_stable([col("B")])
-        .agg([col("A").abs().sum()])
+        .agg([col("A").abs().sum().alias("A_sum")])
         .collect()?;
 
     let out = out.column("A_sum")?;
@@ -2528,36 +1942,6 @@ fn test_apply_flatten() -> Result<()> {
     Ok(())
 }
 
-#[test]
-fn test_agg_unique_first() -> Result<()> {
-    let df = df![
-        "g"=> [1, 1, 2, 2, 3, 4, 1],
-        "v"=> [1, 2, 2, 2, 3, 4, 1],
-    ]?;
-
-    let out = df
-        .lazy()
-        .groupby_stable([col("g")])
-        .agg([
-            col("v").unique().first(),
-            col("v").unique().sort(false).first().alias("true_first"),
-            col("v").unique().list(),
-        ])
-        .collect()?;
-
-    let a = out.column("v_first").unwrap();
-    let a = a.sum::<i32>().unwrap();
-    // can be both because unique does not guarantee order
-    assert!(a == 10 || a == 11);
-
-    let a = out.column("true_first").unwrap();
-    let a = a.sum::<i32>().unwrap();
-    // can be both because unique does not guarantee order
-    assert_eq!(a, 10);
-
-    Ok(())
-}
-
 #[test]
 #[cfg(feature = "is_in")]
 fn test_is_in() -> Result<()> {
